import logging
import warnings
from inspect import cleandoc
from typing import Any, Dict, List, Optional, Set, Union

from pydantic import Field, StringConstraints
from typing_extensions import Annotated

from geolib.geometry import Point
from geolib.models import BaseDataClass
from geolib.models.dseries_parser import (
    DSeriesInlineMappedProperties,
    DSeriesInlineReversedProperties,
    DSeriesNoParseSubStructure,
    DSeriesRepeatedGroupedProperties,
    DSeriesStructure,
    DSeriesStructureCollection,
    DSeriesUnmappedNameProperties,
    DSeriesWrappedTableStructure,
    DSerieVersion,
)
from geolib.utils import make_newline_validator

from .calculation_options import (
    CalculationOptionsPerStage as CalculationOptionsPerStageExternal,
)
from .dsheetpiling_structures import (
    DSeriesPilingDumpParserStructure,
    DSeriesPilingParserStructure,
    DSheetpilingSurchargeLoad,
    DSheetpilingTableEntry,
    DSheetpilingUnwrappedTable,
    DSheetpilingWithNumberOfRowsTable,
)
from .dsheetpiling_validator import DSheetPilingValidator
from .internal_partial_factors import (
    PartialFactorsCurI,
    PartialFactorsCurIi,
    PartialFactorsCurIii,
    PartialFactorsEc7BE1Set1,
    PartialFactorsEc7BE1Set2,
    PartialFactorsEc7BE2Set1,
    PartialFactorsEc7BE2Set2,
    PartialFactorsEc7BE3Set1,
    PartialFactorsEc7BE3Set2,
    PartialFactorsEc7Nl0,
    PartialFactorsEc7Nl1,
    PartialFactorsEc7Nl2,
    PartialFactorsEc7Nl3,
    PartialFactorsEurocodeDa1Set1,
    PartialFactorsEurocodeDa1Set2,
    PartialFactorsEurocodeDa2,
    PartialFactorsEurocodeDa3,
)
from .settings import (
    AssessmentTypeEC7NL,
    CalculationType,
    CurveSettings,
    DesignType,
    DistributionType,
    EarthPressureCoefficients,
    GrainType,
    HorizontalBehaviorType,
    LambdaType,
    LateralEarthPressureMethod,
    LateralEarthPressureMethodStage,
    LoadTypeFavourableUnfavourable,
    LoadTypeFavourableUnfavourableMoment,
    LoadTypePermanentVariable,
    ModelType,
    ModulusReactionType,
    ModulusSubgradeReaction,
    PartialFactorCalculationType,
    PartialFactorSetCUR,
    PartialFactorSetEC,
    PartialFactorSetEC7NADBE,
    PartialFactorSetEC7NADNL,
    PartialFactorSetVerifyEC,
    PassiveSide,
    RiskClassEC7BE,
    SheetPilingElementMaterialType,
    Side,
    SoilTypeModulusSubgradeReaction,
    VerifyType,
)

logger = logging.getLogger(__name__)


_DEFAULT_WATER_LEVEL_NAME: str = "New Water Level"
_DEFAULT_SOIL_PROFILE_NAME: str = "New Profile"
_DEFAULT_SURFACE_NAME: str = "New Surface"
_DEFAULT_SHEET_PILING_ELEMENT_NAME: str = "New element"
_DEFAULT_UNIT_WEIGHT_WATER: float = 9.81
_DEFAULT_PRE_STRESS: float = 0.0

_DEFAULT_SOIL_VERSION: int = 1010
_DEFAULT_SHEETPILING_VERSION: int = 1033

REQ_RUN_LINES = 2


class Model(DSeriesInlineReversedProperties):
    model: ModelType = ModelType.SHEET_PILING
    method: LateralEarthPressureMethod = LateralEarthPressureMethod.MIXED
    verification: bool = False
    pile_load_option: bool = False
    pile_load_by_user: bool = False
    probabilistic: bool = False
    check_vertical_balance: bool = True
    trildens_calculation: bool = True
    diepwand_calculation: bool = False
    elastic_calculation: bool = True
    wooden_sheetpiling: bool = False


class Version(DSerieVersion):
    soil: int = _DEFAULT_SOIL_VERSION
    d__sheet_piling: int = _DEFAULT_SHEETPILING_VERSION


class VersionExternals(DSeriesInlineMappedProperties):
    dgscptipcalc____dll: str = "21.3.1.35380"
    dgscptipui____dll: str = "21.3.1.35380"


class Soil(DSeriesUnmappedNameProperties):
    name: Annotated[str, StringConstraints(min_length=1, max_length=25)] = "default soil"
    soilcolor: int = 9764853
    soilsoiltype: SoilTypeModulusSubgradeReaction = SoilTypeModulusSubgradeReaction.SAND
    soilgraintype: GrainType = GrainType.FINE
    soilgamdry: Annotated[float, Field(ge=0, le=100)] = 0.01
    soilgamwet: Annotated[float, Field(ge=0, le=100)] = 0.01
    soilrelativedensity: Annotated[float, Field(ge=0, le=100)] = 0
    soilemodmenard: Annotated[float, Field(ge=0.01, le=1000000)] = 0.01
    soilcohesion: Annotated[float, Field(ge=0, le=1000000000)] = 0
    soilphi: Annotated[float, Field(ge=-89, le=89)] = 0
    soildelta: Annotated[float, Field(ge=-80, le=89)] = 0.00
    soilisdeltaangleautomaticallycalculated: bool = False
    soilocr: Annotated[float, Field(ge=0, le=1000)] = 1.00
    soilpermeabkx: Annotated[float, Field(ge=0.00000000001, le=1000000)] = 0.0001
    soilstdcohesion: Annotated[float, Field(ge=0, le=100000000)] = 0.00
    soilstdphi: Annotated[float, Field(ge=0, le=100000000)] = 0.00
    soildistcohesion: DistributionType = DistributionType.NONE
    soildistphi: DistributionType = DistributionType.NONE
    soilla: Annotated[float, Field(ge=0, le=1000)] = 0
    soilln: Annotated[float, Field(ge=0, le=1000)] = 0.01
    soillp: Annotated[float, Field(ge=0, le=1000)] = 0.01
    soilusemenard: ModulusSubgradeReaction = ModulusSubgradeReaction.MANUAL
    soilusebrinchhansen: EarthPressureCoefficients = EarthPressureCoefficients.MANUAL
    soilshellfactor: Annotated[float, Field(ge=1, le=1000)] = 1.00
    soillambdatype: LambdaType = LambdaType.MANUAL
    soillam1: Annotated[float, Field(ge=0, le=100)] = 50
    soillam2: Annotated[float, Field(ge=0, le=100)] = 80
    soillam3: Annotated[float, Field(ge=0, le=100)] = 90
    soilkb0: Annotated[float, Field(ge=0.01, le=10000000000)] = 1.00000e04
    soilkb1: Annotated[float, Field(ge=0.01, le=10000000000)] = 5.00000e00
    soilkb2: Annotated[float, Field(ge=0.01, le=10000000000)] = 1.00000e04
    soilkb3: Annotated[float, Field(ge=0.01, le=10000000000)] = 1.00000e04
    soilkb4: Annotated[float, Field(ge=0.01, le=10000000000)] = 1.00000e04
    soilko0: Annotated[float, Field(ge=0.01, le=10000000000)] = 1.00000e04
    soilko1: Annotated[float, Field(ge=0.01, le=10000000000)] = 5.00000e00
    soilko2: Annotated[float, Field(ge=0.01, le=10000000000)] = 1.00000e04
    soilko3: Annotated[float, Field(ge=0.01, le=10000000000)] = 1.00000e04
    soilko4: Annotated[float, Field(ge=0.01, le=10000000000)] = 1.00000e04
    soilcurkb1: Annotated[float, Field(ge=0.01, le=10000000000)] = 2.00000e03
    soilcurkb2: Annotated[float, Field(ge=0.01, le=10000000000)] = 8.00000e02
    soilcurkb3: Annotated[float, Field(ge=0.01, le=10000000000)] = 5.00000e02
    soilcurko1: Annotated[float, Field(ge=0.01, le=10000000000)] = 2.00000e03
    soilcurko2: Annotated[float, Field(ge=0.01, le=10000000000)] = 8.00000e02
    soilcurko3: Annotated[float, Field(ge=0.01, le=10000000000)] = 5.00000e02
    soilhorizontalbehaviourtype: HorizontalBehaviorType = HorizontalBehaviorType.ELASTIC
    soilelasticity: float = 1.00000e03  # fixed value
    soildefaultelasticity: int = 1  # fixed value


class SoilCollection(DSeriesStructureCollection):
    soil: List[Soil] = []

    @property
    def soil_names(self) -> Set[str]:
        return {soil.name for soil in self.soil}

    def add_soil_if_unique(self, soil) -> None:
        for added_soil in self.soil:
            if soil.name == added_soil.name:
                raise NameError(f"Soil with name {soil.name} already exists.")
        self.soil.append(soil)
        return soil

    def find_soil_id(self, key) -> int:
        for i, soil in enumerate(self.soil):
            if soil.name == key:
                return i
        raise KeyError(f"Soil with {key} not present.")

    def __getitem__(self, key) -> Soil:
        if isinstance(key, int):
            return self.soil[key]
        elif isinstance(key, str):
            for soil in self.soil:
                if soil.name == key:
                    return soil
            raise KeyError(key)


class SoilLayer(DSeriesNoParseSubStructure):
    top_of_layer: float
    soil: str
    water_pressure_top: float
    water_pressure_bottom: float


class SoilProfile(DSeriesNoParseSubStructure):
    name: Annotated[str, StringConstraints(min_length=1, max_length=25)]
    layers: List[SoilLayer]
    coordinate: Point


class SoilProfiles(DSeriesNoParseSubStructure):
    soil_profiles: List[SoilProfile] = []
    curve_number: Annotated[int, Field(ge=1, le=4)] = 3
    use_unloading_reloading_curve: bool = False
    modulus_reaction_type: int = ModulusReactionType.SECANT.value

    @property
    def soil_profile_names(self) -> Set[str]:
        return {soil_profile.name for soil_profile in self.soil_profiles}

    @property
    def referenced_soil_names(self) -> Set[str]:
        return {layer.soil for profile in self.soil_profiles for layer in profile.layers}


class CalculationOptions(DSeriesStructure):
    # TODO these first values are not used for single pile
    calcfirststageinitial: bool = False
    calcminornodeson: bool = False
    calcreducedeltas: bool = False
    inputcalculationtype: CalculationType = CalculationType.STANDARD
    isvibrationcalculation: bool = False
    allowableanchorforcecalculationtype: bool = False
    # available when c, phi , delta is selected for as a model
    calcautolambdason: bool = True
    # design sheet pile length calculation
    designstage: Annotated[int, Field(ge=0)] = 0
    designpilelengthfrom: Annotated[float, Field(ge=1, le=100)] = 1
    designpilelengthto: Annotated[float, Field(ge=1, le=100)] = 1
    designpilelengthdecrement: Annotated[float, Field(ge=0.01, le=10)] = 0.01
    designpilelengthnew: int = 1  # fixed value
    designtype: DesignType = DesignType.REPRESENTATIVE
    designeurocodepartialfactorset: PartialFactorSetEC = PartialFactorSetEC.DA1SET1
    designpartialfactorsetec7nadnl: PartialFactorSetEC7NADNL = (
        PartialFactorSetEC7NADNL.RC0
    )
    designec7nlmethod: PartialFactorCalculationType = PartialFactorCalculationType.METHODA
    designec7bmethod: PartialFactorCalculationType = PartialFactorCalculationType.METHODA
    designec7bepartialfactorset: PartialFactorSetEC7NADBE = (
        PartialFactorSetEC7NADBE.RC1SET1
    )
    designec7bemethod: PartialFactorCalculationType = PartialFactorCalculationType.METHODA
    designpartialfactorset: PartialFactorSetCUR = PartialFactorSetCUR.CLASSI
    designcurmethod: PartialFactorCalculationType = PartialFactorCalculationType.METHODA
    # verify sheet piling calculation
    verifytype: VerifyType = VerifyType.CUR
    eurocodepartialfactorset: PartialFactorSetVerifyEC = PartialFactorSetVerifyEC.DA1
    eurocodeoverallstability: bool = False
    ec7nlmethod: PartialFactorCalculationType = PartialFactorCalculationType.METHODA
    ec7nloverallpartialfactorset: PartialFactorSetEC7NADNL = PartialFactorSetEC7NADNL.RC0
    ec7nloverallanchorfactor: Annotated[float, Field(ge=0.001, le=1000)] = 1
    ec7nadnloverallstability: bool = False
    ec7nlassessmenttype: AssessmentTypeEC7NL = AssessmentTypeEC7NL.NewConstruction
    ec7beoverallstability: bool = False
    ec7bemethod: PartialFactorCalculationType = PartialFactorCalculationType.METHODA
    ec7beoverallriskclass: RiskClassEC7BE = RiskClassEC7BE.RC2
    nbmethod: PartialFactorCalculationType = PartialFactorCalculationType.METHODA
    curmethod: PartialFactorCalculationType = PartialFactorCalculationType.METHODA
    curoverallpartialfactorset: PartialFactorSetCUR = PartialFactorSetCUR.CLASSI
    curoverallanchorfactor: Annotated[float, Field(ge=0.001, le=1000)] = 1
    curoverallstability: bool = False
    # Characteristic Kranz Anchor Strength calculation
    curanchorforcestage: Annotated[int, Field(ge=0)] = 0
    # Overall stability calculation
    curstabilitystage: Annotated[int, Field(ge=0)] = 0
    overallstabilitytype: DesignType = DesignType.REPRESENTATIVE
    stabilityexport: bool = False
    stabilityeurocodepartialfactorset: PartialFactorSetEC = PartialFactorSetEC.DA1SET1
    stabilityec7nlpartialfactorset: PartialFactorSetEC7NADNL = (
        PartialFactorSetEC7NADNL.RC0
    )
    overallstabilityec7bepartialfactorset: PartialFactorSetEC7NADBE = (
        PartialFactorSetEC7NADBE.RC1SET1
    )
    stabilitycurpartialfactorset: PartialFactorSetCUR = PartialFactorSetCUR.CLASSI

    # These are all subgroups (key=value)
    partial_factors_eurocode_da1_set1: PartialFactorsEurocodeDa1Set1 = (
        PartialFactorsEurocodeDa1Set1()
    )
    partial_factors_eurocode_da1_set2: PartialFactorsEurocodeDa1Set2 = (
        PartialFactorsEurocodeDa1Set2()
    )
    partial_factors_eurocode_da2: PartialFactorsEurocodeDa2 = PartialFactorsEurocodeDa2()
    partial_factors_eurocode_da3: PartialFactorsEurocodeDa3 = PartialFactorsEurocodeDa3()
    partial_factors_ec7_nl_0: PartialFactorsEc7Nl0 = PartialFactorsEc7Nl0()
    partial_factors_ec7_nl_1: PartialFactorsEc7Nl1 = PartialFactorsEc7Nl1()
    partial_factors_ec7_nl_2: PartialFactorsEc7Nl2 = PartialFactorsEc7Nl2()
    partial_factors_ec7_nl_3: PartialFactorsEc7Nl3 = PartialFactorsEc7Nl3()
    partial_factors_ec7_be_1_set1: PartialFactorsEc7BE1Set1 = PartialFactorsEc7BE1Set1()
    partial_factors_ec7_be_1_set2: PartialFactorsEc7BE1Set2 = PartialFactorsEc7BE1Set2()
    partial_factors_ec7_be_2_set1: PartialFactorsEc7BE2Set1 = PartialFactorsEc7BE2Set1()
    partial_factors_ec7_be_2_set2: PartialFactorsEc7BE2Set2 = PartialFactorsEc7BE2Set2()
    partial_factors_ec7_be_3_set1: PartialFactorsEc7BE3Set1 = PartialFactorsEc7BE3Set1()
    partial_factors_ec7_be_3_set2: PartialFactorsEc7BE3Set2 = PartialFactorsEc7BE3Set2()
    partial_factors_cur_i: PartialFactorsCurI = PartialFactorsCurI()
    partial_factors_cur_ii: PartialFactorsCurIi = PartialFactorsCurIi()
    partial_factors_cur_iii: PartialFactorsCurIii = PartialFactorsCurIii()


class SheetPileElement(DSeriesUnmappedNameProperties):
    name: Annotated[
        str, StringConstraints(min_length=1, max_length=50)
    ] = _DEFAULT_SHEET_PILING_ELEMENT_NAME
    sheetpilingelementmaterialtype: SheetPilingElementMaterialType = (
        SheetPilingElementMaterialType.Steel
    )
    sheetpilingelementei: Annotated[float, Field(ge=0.001, le=1e12)] = 100000
    sheetpilingelementwidth: Annotated[float, Field(ge=0, le=1000)] = 1
    sheetpilingelementlevel: Annotated[float, Field(ge=-10000, le=10000)] = -10
    sheetpilingelementheight: Annotated[int, Field(ge=10, le=100000)] = 400
    sheetpilingpilewidth: Annotated[float, Field(ge=0, le=100000)] = 0
    sheetpilingelementsectionarea: Annotated[int, Field(ge=10, le=100000)] = 170
    sheetpilingelementresistingmoment: Annotated[int, Field(ge=0, le=100000)] = 0
    sheetpilingelementreductionfactorei: Annotated[float, Field(ge=0.01, le=10)] = 1
    sheetpilingelementnote: Annotated[
        str, StringConstraints(min_length=0, max_length=20)
    ] = ""
    sheetpilingelementmaxcharacteristicmoment: Annotated[
        float, Field(ge=0, le=100000)
    ] = 0
    sheetpilingelementmaxplasticcharacteristicmoment: Annotated[
        float, Field(ge=0, le=100000)
    ] = 0
    sheetpilingelementkmod: Annotated[float, Field(ge=0.01, le=1)] = 0.01
    sheetpilingelementmaterialfactor: Annotated[float, Field(ge=0.01, le=10)] = 0.01
    ssheetpilingelementreductionfactormaxmoment: Annotated[
        float, Field(ge=0.01, le=10)
    ] = 1
    diaphragmwallissymmetric: Annotated[int, Field(ge=0, le=1)] = 0
    diaphragmwallposeielastoplastic1: Annotated[float, Field(ge=0, le=100000)] = 0
    diaphragmwallnegeielastoplastic1: Annotated[float, Field(ge=0, le=100000)] = 0
    diaphragmwallposmomelastic: Annotated[float, Field(ge=0, le=1000000000000)] = 100000
    diaphragmwallnegmomelastic: Annotated[float, Field(ge=0, le=100000)] = 0
    diaphragmwallposmomplastic: Annotated[float, Field(ge=0, le=100000)] = 0
    diaphragmwallnegmomplastic: Annotated[float, Field(ge=0, le=100000)] = 0
    diaphragmwallposeielastoplastic2: Annotated[float, Field(ge=0, le=100000)] = 0
    diaphragmwallposmomelastoplastic: Annotated[float, Field(ge=0, le=100000)] = 0
    diaphragmwallnegeielastoplastic2: Annotated[float, Field(ge=0, le=100000)] = 0
    diaphragmwallnegmomelastoplastic: Annotated[float, Field(ge=0, le=100000)] = 0
    woodensheetpilingelemente: Annotated[
        float, Field(ge=0.001, le=1000000000000)
    ] = 100000
    woodensheetpilingelementcharacflexuralstrength: Annotated[
        float, Field(ge=0, le=100000)
    ] = 0
    woodensheetpilingelementksys: Annotated[float, Field(ge=0.01, le=10)] = 1.15
    woodensheetpilingelementkdef: Annotated[float, Field(ge=0.01, le=10)] = 1
    woodensheetpilingelementpsi2eff: Annotated[float, Field(ge=0, le=10)] = 1
    woodensheetpilingelementmaterialfactor: Annotated[float, Field(ge=0.01, le=10)] = 1.3
    woodensheetpilingelementkmodfshort: Annotated[float, Field(ge=0.01, le=10)] = 0.65
    woodensheetpilingelementkmodflong: Annotated[float, Field(ge=0.01, le=10)] = 0.5
    woodensheetpilingelementkmode: Annotated[float, Field(ge=0.01, le=10)] = 0.8

    def __init__(self, *args, **kwargs) -> "SheetPileElement":
        default_dict = dict(self.model_fields)
        for field, value in kwargs.items():
            if value is None:
                kwargs[field] = default_dict[field].default
        super().__init__(*args, **kwargs)

    def update_arguments_if_invalid_input(self) -> None:
        """This function updates sheetpilingelementkmod and
        result_dict_gamma_m_value according to the
        sheetpilingelementmaterialtype inputed by the user.

        For some types the user can define their own parameters and for
        some others default values are defined.
        """
        # k-mod value
        result_dict_k_mod_value = {
            0: self.sheetpilingelementkmod,
            1: 1.0,
            2: 1.0,
            3: 1.0,
            4: self.sheetpilingelementkmod,
            5: self.sheetpilingelementkmod,
        }
        if (
            self.sheetpilingelementkmod
            != result_dict_k_mod_value[self.sheetpilingelementmaterialtype.value]
        ):
            self.sheetpilingelementkmod = result_dict_k_mod_value[
                self.sheetpilingelementmaterialtype.value
            ]
            warnings.warn(
                "Inputed value for kmod was not valid and was change to "
                + str(self.sheetpilingelementkmod),
                ImportWarning,
            )
        # gammma_m value
        result_dict_gamma_m_value = {
            0: self.sheetpilingelementmaterialfactor,
            1: 1.0,
            2: 1.1,
            3: 1.0,
            4: 1.2,
            5: self.sheetpilingelementmaterialfactor,
        }
        if (
            self.sheetpilingelementmaterialfactor
            != result_dict_gamma_m_value[self.sheetpilingelementmaterialtype.value]
        ):
            self.sheetpilingelementmaterialfactor = result_dict_gamma_m_value[
                self.sheetpilingelementmaterialtype.value
            ]
            warnings.warn(
                "Inputed value for material factor was not valid and was change to "
                + str(self.sheetpilingelementmaterialfactor),
                ImportWarning,
            )


class SheetPiling(DSeriesStructureCollection):
    sheetpiling: Annotated[List[SheetPileElement], Field(min_length=1)] = [
        SheetPileElement()
    ]
    leveltopsheetpiling: Annotated[float, Field(ge=-10000, le=10000)] = 0.0
    lengthsheetpiling: Annotated[float, Field(gt=0)] = 10


class VerticalBalance(DSeriesInlineMappedProperties):
    sheetpilingqcrep: Optional[Annotated[float, Field(ge=0)]] = 0.001
    sheetpilingxi: Optional[Annotated[float, Field(ge=0.1)]] = 1.39


class Anchor(DSheetpilingTableEntry):
    name: Annotated[str, StringConstraints(min_length=1, max_length=50)]
    level: float = 0
    e_modulus: Annotated[float, Field(gt=0)] = 2.1e8
    cross_section: Optional[Annotated[float, Field(gt=0)]] = 1e-3
    wall_height_kranz: Optional[Annotated[float, Field(ge=0)]] = 0.00
    length: Optional[Annotated[float, Field(gt=0)]] = 1
    angle: Optional[float] = 0.00
    yield_force: Optional[Annotated[float, Field(ge=0)]] = 0.00
    side: Side = Side.RIGHT


class Anchors(DSheetpilingUnwrappedTable):
    """Container for Anchor."""

    anchors: List[Anchor] = []

    @property
    def anchor_names(self) -> Set[str]:
        return {anchor.name for anchor in self.anchors}


class AnchorOrStrutPresstressReference(BaseDataClass):
    """Used in the ConstructionStage to identify anchor and assign a
    prestress."""

    name: str
    pre_stress: Annotated[float, Field(ge=0)] = _DEFAULT_PRE_STRESS


class Strut(DSheetpilingTableEntry):
    name: Annotated[str, StringConstraints(min_length=1, max_length=50)]
    level: float = 0
    e_modulus: Annotated[float, Field(gt=0)] = 2.1e8
    cross_section: Optional[Annotated[float, Field(gt=0)]] = 1e-4
    length: Optional[Annotated[float, Field(gt=0)]] = 1
    angle: Optional[float] = 0.00
    buckling_force: Optional[Annotated[float, Field(ge=0)]] = 0.00
    side: Side = Side.RIGHT


class Struts(DSheetpilingUnwrappedTable):
    """Container for Strut."""

    struts: List[Strut] = []

    @property
    def strut_names(self) -> Set[str]:
        return {strut.name for strut in self.struts}


class ConstructionStage(DSeriesUnmappedNameProperties):
    name: Annotated[str, StringConstraints(min_length=1, max_length=50)]
    displacement_top_sheet_piling: float = 0
    passive_side: PassiveSide = PassiveSide.DSHEETPILING_DETERMINED
    method_left: LateralEarthPressureMethodStage = (
        LateralEarthPressureMethodStage.KA_KO_KP
    )
    method_right: LateralEarthPressureMethodStage = (
        LateralEarthPressureMethodStage.KA_KO_KP
    )
    water_level_left: str = _DEFAULT_WATER_LEVEL_NAME
    water_level_right: str = _DEFAULT_WATER_LEVEL_NAME
    surface_left: str = _DEFAULT_SURFACE_NAME
    surface_right: str = _DEFAULT_SURFACE_NAME
    soil_profile_left: Optional[str] = _DEFAULT_SOIL_PROFILE_NAME
    soil_profile_right: Optional[str] = _DEFAULT_SOIL_PROFILE_NAME
    anchors: List[AnchorOrStrutPresstressReference] = []
    struts: List[AnchorOrStrutPresstressReference] = []
    spring_supports: List[str] = []
    rigid_supports: List[str] = []
    uniform_loads: List[str] = []
    surcharge_loads_left: List[str] = []
    surcharge_loads_right: List[str] = []
    horizontal_line_loads: List[str] = []
    moment_loads: List[str] = []
    normal_forces: List[str] = []


class ConstructionStages(DSeriesStructureCollection):
    stages: List[ConstructionStage] = []

    @property
    def stage_names(self) -> Set[str]:
        return {stage.name for stage in self.stages}


class WaterLevel(DSeriesNoParseSubStructure):
    name: Annotated[str, StringConstraints(min_length=1, max_length=50)]
    level: float
    distribution_type: DistributionType = DistributionType.NONE
    standard_deviation: float = 0.0


class WaterLevels(DSeriesNoParseSubStructure):
    levels: List[WaterLevel] = []

    @property
    def water_level_names(self) -> Set[str]:
        return {water_level.name for water_level in self.levels}


class StageOptions(DSeriesInlineMappedProperties):
    """Representation of [STAGE] block."""

    stagepartialfactorsetcur: PartialFactorSetCUR = PartialFactorSetCUR.UNKNOWN
    stageverify: int = 0
    stageanchorfactor: Annotated[float, Field(ge=0.001, le=1000)] = 1
    stagepartialfactorsetec7nadnl: PartialFactorSetEC7NADNL = PartialFactorSetEC7NADNL.RC0
    stageriskclassec7nadbe: RiskClassEC7BE = RiskClassEC7BE.RC2
    stageverifyec7nadnl: int = 0
    stageanchorfactorec7nadnl: Annotated[float, Field(ge=0.001, le=1000)] = 1
    stageverifyec7nadbe: int = 0


class CalculationOptionsPerStage(DSeriesStructureCollection):
    """Representation of [CALCULATION OPTIONS PER STAGE] block."""

    stageoptions: List[StageOptions] = []


class UniformLoad(DSeriesUnmappedNameProperties):
    name: Annotated[str, StringConstraints(min_length=1, max_length=50)]
    uniformloadleft: float
    uniformloadright: float
    uniformloadpermanent: LoadTypePermanentVariable = LoadTypePermanentVariable.PERMANENT
    uniformloadfavourable: LoadTypeFavourableUnfavourable = (
        LoadTypeFavourableUnfavourable.DSHEETPILING_DETERMINED
    )
    uniformloadstandarddeviationleft: float
    uniformloadstandarddeviationright: float
    uniformloaddistleft: DistributionType = DistributionType.NORMAL
    uniformloaddistright: DistributionType = DistributionType.NORMAL


class UniformLoads(DSeriesStructureCollection):
    loads: List[UniformLoad] = []

    @property
    def load_names(self) -> Set[str]:
        return {load.name for load in self.loads}


class SurchargePoint(DSeriesInlineMappedProperties):
    surchargeloaddistance: float
    surchargeloadvalue: float


class SurchargeLoad(DSheetpilingSurchargeLoad):
    name: Annotated[str, StringConstraints(min_length=1, max_length=50)]
    points: List[SurchargePoint] = []
    surchargeloadpermanent: LoadTypePermanentVariable = (
        LoadTypePermanentVariable.PERMANENT
    )
    surchargeloadfavourable: LoadTypeFavourableUnfavourable = (
        LoadTypeFavourableUnfavourable.DSHEETPILING_DETERMINED
    )
    surchargeloadstandarddeviation: float = 0.00
    surchargeloaddistribution: DistributionType = DistributionType.NORMAL


class SurchargeLoads(DSeriesStructureCollection):
    loads: List[SurchargeLoad] = []

    @property
    def load_names(self) -> Set[str]:
        return {load.name for load in self.loads}


class Surface(DSeriesNoParseSubStructure):  # TODO determine structure
    name: Annotated[str, StringConstraints(min_length=1, max_length=50)]
    points: Annotated[List[Point], Field(min_length=1)]
    points: Annotated[List[dict], Field(min_length=1)]
    distribution_type: DistributionType = DistributionType.NONE
    std: Annotated[float, Field(ge=0.0)] = 0.0


class Surfaces(DSeriesNoParseSubStructure):  # TODO GroupList should be suitable?
    surfaces: List[Surface] = []

    @property
    def surface_names(self) -> Set[str]:
        return {surface.name for surface in self.surfaces}


class HorizontalLineLoad(DSeriesNoParseSubStructure):
    name: Annotated[str, StringConstraints(min_length=1, max_length=50)]
    level: Annotated[float, Field(ge=-1e12, le=1e12)] = 0
    load: Annotated[float, Field(ge=-1e12, le=1e12)] = 100
    load_type: LoadTypeFavourableUnfavourable = (
        LoadTypeFavourableUnfavourable.DSHEETPILING_DETERMINED
    )
    duration_type: LoadTypePermanentVariable = LoadTypePermanentVariable.PERMANENT


class HorizontalLineLoads(DSeriesNoParseSubStructure):
    loads: List[HorizontalLineLoad] = []


class Moment(DSeriesNoParseSubStructure):
    name: Annotated[str, StringConstraints(min_length=1, max_length=50)]
    level: Annotated[float, Field(ge=-1e12, le=1e12)] = 0
    load: Annotated[float, Field(ge=-1e12, le=1e12)] = 100
    load_type: LoadTypeFavourableUnfavourableMoment = (
        LoadTypeFavourableUnfavourableMoment.FAVOURABLE
    )
    duration_type: LoadTypePermanentVariable = LoadTypePermanentVariable.PERMANENT


class Moments(DSeriesNoParseSubStructure):
    loads: List[Moment] = []


class NormalForce(DSeriesNoParseSubStructure):
    name: Annotated[str, StringConstraints(min_length=1, max_length=50)]
    force_at_sheet_pile_top: Annotated[float, Field(ge=-1e12, le=1e12)]
    force_at_surface_level_left_side: Annotated[float, Field(ge=-1e12, le=1e12)]
    force_at_surface_level_right_side: Annotated[float, Field(ge=-1e12, le=1e12)]
    force_at_sheet_pile_toe: Annotated[float, Field(ge=-1e12, le=1e12)]
    load_type: LoadTypeFavourableUnfavourableMoment = (
        LoadTypeFavourableUnfavourableMoment.FAVOURABLE
    )
    duration_type: LoadTypePermanentVariable = LoadTypePermanentVariable.PERMANENT


class NormalForces(DSeriesNoParseSubStructure):
    loads: List[NormalForce] = []


class Support(DSeriesNoParseSubStructure):
    """Internal structure for spring and rigid supports."""

    name: Annotated[str, StringConstraints(min_length=1, max_length=50)]
    level: float
    rotational_stiffness: Annotated[float, Field(ge=0)]
    translational_stiffness: Annotated[float, Field(ge=0)]


class SupportContainer(DSeriesNoParseSubStructure):
    supports: List[Support] = []

    @property
    def support_names(self) -> Set[str]:
        return {support.name for support in self.supports}


class VibrationPosition(DSeriesNoParseSubStructure):
    x: float
    z: float


class VibrationPositions(DSeriesNoParseSubStructure):
    positions: List[VibrationPosition] = []


class Water(DSeriesUnmappedNameProperties):
    unit_weight_of_water: float = 9.81


class DSheetPilingInputStructure(DSeriesStructure):
    """Representation of complete .shi file."""

    version: Version = Version()
    version_externals: VersionExternals = VersionExternals()
    soil_collection: SoilCollection = SoilCollection()
    run_identification: str = 2 * "\n"
    model: Model = Model()
    cpt_list: str = cleandoc(
        """
        Count=0
        """
    )
    sheet_piling: Union[str, SheetPiling] = SheetPiling()
    combined_wall: str = ""
    vertical_balance: VerticalBalance = VerticalBalance()
    settlement_by_vibration_params: str = cleandoc(
        """
        SheetPilingNumberOfPilesDrilled=2
        """
    )
    horizontal_line_loads: Union[HorizontalLineLoads, str, None] = None
    uniform_loads: Optional[UniformLoads] = None
    surcharge_loads: Optional[SurchargeLoads] = None
    water: str = ""
    earth_quake: str = cleandoc(
        """
        0.00
        """
    )
    soil_profiles: Union[SoilProfiles, str] = cleandoc(
        f"""
          1 Number of spring characteristics curves
          0 1/0 : Yes/No Unloading curve
          1 1/0 : Yes/No Use Cur values

          1 Number of soil profiles 
        {_DEFAULT_SOIL_PROFILE_NAME}
               0.000 X coordinate
               0.000 Y coordinate

          1 Number of soil layers per soil profile
         Nr      Level   Wosp Top  Wosp Bott Material Name
          1       0.00       0.00       0.00 New Material

        """
    )
    surfaces: Union[str, Surfaces] = cleandoc(
        f"""
        1 Number of surfaces 
        1     1 {_DEFAULT_SURFACE_NAME}
        0.00  Standard deviation
        2  Distribution type
        Nr     X-coord     Value
        1       0.00       0.00
        """
    )
    water: Water = Water()
    waterlevels: Union[WaterLevels, str] = cleandoc(
        f"""
          1 Number of Waterlevels 
          3 Number of Data per Waterlevel 
        {_DEFAULT_WATER_LEVEL_NAME}
              0.00
              0.00
                 2

        """
    )
    construction_stages: Union[str, ConstructionStages] = ConstructionStages()
    calculation_options_per_stage: CalculationOptionsPerStage = (
        CalculationOptionsPerStage()
    )
    calculation_options: CalculationOptions = CalculationOptions()
    probabilistic_calculation_options: str = cleandoc(
        """
        ZFunctionType=1
        ZFunctionMobilisation=100.00
        ZFunctionMoment=0.00
        ZFunctionAnchorForce=0.00
        ZFunctionAnchor=-1
        """
    )
    probabilistic_control_parameters: str = cleandoc(
        """
        IterationFactor=1.000
        IterationNumber=15
        CriteriaLimit1=0.010000
        CriteriaLimit2=0.010000
        ReliabilityStage=0
        """
    )

    """
    When there are no anchors, struts, supports, vibration positions or soil displacements in the model,
    their respective block is not present in the .shi file.
    """
    soil_displacements: Optional[str] = None
    rigid_supports: Union[str, SupportContainer, None] = None
    spring_supports: Union[str, SupportContainer, None] = None
    moments: Union[str, Moments, None] = None
    normal_forces: Union[str, NormalForces, None] = None
    anchors: Union[str, Anchors, None] = None
    struts: Union[str, Struts, None] = None
    vibration_positions: Union[str, VibrationPositions, None] = None

    # Custom validator
    _validate_run_identification = make_newline_validator(
        "run_identification", req_newlines=REQ_RUN_LINES
    )

    @property
    def is_valid(self) -> bool:
        return self._validator().is_valid

    def has_stage(self, stage_id: int) -> bool:
        try:
            self.construction_stages.stages[stage_id]
            return True
        except IndexError:
            return False

    def _validator(self) -> DSheetPilingValidator:
        return DSheetPilingValidator(self)

    def set_model(self, **kwargs) -> None:
        kwargs = self._filter_none_values_from_key_value_dict(kwargs)
        self.model = Model(**kwargs)

        # TODO Singlepile model is a won't have, raise error here?

        if self.model.method != LateralEarthPressureMethod.MIXED:
            for stage in self.construction_stages.stages:
                stage.method_left = (
                    LateralEarthPressureMethodStage.get_stage_type_from_method(
                        self.model.method
                    )
                )
                stage.method_right = (
                    LateralEarthPressureMethodStage.get_stage_type_from_method(
                        self.model.method
                    )
                )

    def set_calculation_options(self, **kwargs) -> None:
        kwargs = self._filter_none_values_from_key_value_dict(kwargs)
        # here add function that turns kwargs from snake case to pascal case
        kwargs = self._from_snake_to_pascal_case(kwargs)
        self.calculation_options = CalculationOptions(**kwargs)

    def add_calculation_options_per_stage(
        self,
        input_calc_options: CalculationOptionsPerStageExternal,
        stage_id: int,
    ) -> None:
        _map_external_to_internal_values = {
            VerifyType.CUR: {
                "stagepartialfactorsetcur": input_calc_options.partial_factor_set,
                "stageverify": stage_id + 1,
                "stageanchorfactor": input_calc_options.anchor_factor,
            },
            VerifyType.EC7NL: {
                "stagepartialfactorsetec7nadnl": input_calc_options.partial_factor_set,
                "stageverifyec7nadnl": stage_id + 1,
                "stageanchorfactorec7nadnl": input_calc_options.anchor_factor,
            },
            VerifyType.EC7BE: {
                "stageverifyec7nadbe": stage_id + 1,
            },
        }
        stageoptions = StageOptions(
            **_map_external_to_internal_values[self.calculation_options.verifytype]
        )
        self.calculation_options_per_stage.stageoptions[stage_id] = stageoptions

    def set_curve_settings(self, curve_settings: CurveSettings) -> None:
        if isinstance(self.soil_profiles, str):
            self.soil_profiles = SoilProfiles()
        self.soil_profiles.curve_number = curve_settings.curve_number
        self.soil_profiles.use_unloading_reloading_curve = (
            curve_settings.use_unloading_reloading_curve
        )
        self.soil_profiles.modulus_reaction_type = (
            curve_settings.modulus_reaction_type.value
        )

    def add_stage(
        self,
        name: str,
        passive_side: PassiveSide,
        method_left: LateralEarthPressureMethodStage,
        method_right: LateralEarthPressureMethodStage,
        pile_top_displacement: Optional[float],
    ) -> None:
        if isinstance(self.construction_stages, str):
            self.construction_stages = ConstructionStages()
        if name in self.construction_stages.stage_names:
            raise ValueError(
                f"Stage name {name} already present: all stage names must be unique"
            )

        method_left = self._get_validated_lateral_earth_pressure_method(method_left)
        method_right = self._get_validated_lateral_earth_pressure_method(method_right)
        self._validate_lateral_earth_pressure_methods_for_single_pile_model(
            method_left, method_right
        )

        kwargs = self._filter_none_values_from_key_value_dict(
            {
                "name": name,
                "displacement_top_sheet_piling": pile_top_displacement,
                "passive_side": passive_side,
                "method_left": method_left,
                "method_right": method_right,
            }
        )
        self.construction_stages.stages.append(ConstructionStage(**kwargs))
        # defaults for calculation options should be initialised here
        self.calculation_options_per_stage.stageoptions.append(StageOptions())

    def _get_validated_lateral_earth_pressure_method(
        self, stage_method: Optional[LateralEarthPressureMethodStage]
    ) -> LateralEarthPressureMethodStage:
        """When no stage_method is provided, returns the method set in the
        model options.

        Raises ValueError if method is not compatible with method set on
        model.
        """
        if not isinstance(self.model, Model):
            raise ValueError("Model is a string; set model first")
        if stage_method is None:
            return self.model.method
        error = ValueError(
            f"stage_method {stage_method} is not compatible with method on model {self.model.method}"
        )
        if (
            self.model.method.name == LateralEarthPressureMethodStage.KA_KO_KP.name
            and stage_method.name == LateralEarthPressureMethodStage.C_PHI_DELTA.name
        ):
            raise error
        if (
            self.model.method.name == LateralEarthPressureMethodStage.C_PHI_DELTA.name
            and stage_method.name == LateralEarthPressureMethodStage.KA_KO_KP.name
        ):
            raise error
        return stage_method

    def _validate_lateral_earth_pressure_methods_for_single_pile_model(
        self,
        method_left: LateralEarthPressureMethodStage,
        method_right: LateralEarthPressureMethodStage,
    ) -> None:
        if not isinstance(self.model, Model):
            raise ValueError("Model is a string; set model first")
        if self.model.model == ModelType.SINGLE_PILE:
            if method_left != method_right:
                raise ValueError(
                    f"For the Single Pile Model the left and right method must be the same. Received: left {method_left} and right {method_right}"
                )

    def _filter_none_values_from_key_value_dict(self, dict_: Dict) -> Dict:
        # This way defaults can be defined only in the internal.
        return {k: v for k, v in dict_.items() if v is not None}

    def _from_snake_to_pascal_case(self, dict_: Dict) -> Dict:
        return {k.replace("_", ""): v for k, v in dict_.items()}

    def add_water_level(self, stage_id: int, water_level: WaterLevel, side: Side) -> None:
        # Add water level to water levels
        if isinstance(self.waterlevels, str):
            self.waterlevels = WaterLevels()
        if water_level.name not in self.waterlevels.water_level_names:
            self.waterlevels.levels.append(water_level)

        # Add water level to stages.
        stage = self.construction_stages.stages[stage_id]
        if side == Side.BOTH:
            stage.water_level_left = stage.water_level_right = water_level.name
        elif side == Side.LEFT:
            stage.water_level_left = water_level.name
        elif side == Side.RIGHT:
            stage.water_level_right = water_level.name
        else:
            raise ValueError(f"Provide a Side, received {side}")

    def add_profile(self, stage_id: int, profile: SoilProfile, side: Side) -> None:
        # Add profile to profiles
        if isinstance(self.soil_profiles, str):
            self.soil_profiles = SoilProfiles()
        if profile.name not in self.soil_profiles.soil_profile_names:
            self.soil_profiles.soil_profiles.append(profile)

        soil_names = self.soil_collection.soil_names
        for i, layer in enumerate(profile.layers):
            if layer.soil not in soil_names:
                raise ValueError(
                    f"Soil {layer.soil} of layer {i} is not in soil collection; add soil first via model."
                )
        # Add profile to stages.
        if side == Side.BOTH:
            self.construction_stages.stages[stage_id].soil_profile_left = profile.name
            self.construction_stages.stages[stage_id].soil_profile_right = profile.name
        elif side == Side.LEFT:
            self.construction_stages.stages[stage_id].soil_profile_left = profile.name
        elif side == Side.RIGHT:
            self.construction_stages.stages[stage_id].soil_profile_right = profile.name
        else:
            raise ValueError(f"Provide a Side, received {side}")

    def add_surface(self, stage_id: int, surface: Surface, side: Side) -> None:
        # Add surface to surfaces
        if isinstance(self.surfaces, str):
            self.surfaces = Surfaces()
        if surface.name not in self.surfaces.surface_names:
            self.surfaces.surfaces.append(surface)

        # Add surface to stages.
        if side == Side.BOTH:
            self.construction_stages.stages[stage_id].surface_left = surface.name
            self.construction_stages.stages[stage_id].surface_right = surface.name
        elif side == Side.LEFT:
            self.construction_stages.stages[stage_id].surface_left = surface.name
        elif side == Side.RIGHT:
            self.construction_stages.stages[stage_id].surface_right = surface.name
        else:
            raise ValueError(f"Provide a Side, received {side}")

    def set_construction(
        self, top_level: float, elements: List[SheetPileElement]
    ) -> None:
        elements.sort(key=lambda element: element.sheetpilingelementlevel, reverse=True)
        for sheet in elements:
            sheet.update_arguments_if_invalid_input()
        try:
            bottom_level = elements[-1].sheetpilingelementlevel
        except IndexError:
            raise ValueError("At least one sheet should be provided")

        self.sheet_piling = SheetPiling(
            sheetpiling=elements,
            leveltopsheetpiling=top_level,
            lengthsheetpiling=top_level - bottom_level,
        )

    def add_element_in_sheet_piling(
        self,
        sheet: Any,
        location_top: Optional[Point] = None,
    ) -> None:
        self.sheet_piling.update_level_top_sheet_pile(location_top)
        try:
            sheet = sheet._to_internal()
        except AttributeError:
            raise ValueError(f"Invalid argument, expected sheet, got {type(sheet)}")
        sheet.update_arguments_if_invalid_input()
        self.sheet_piling.sheetpiling.append(sheet)
        self.sheet_piling.update_length_of_sheet_pile()

    def set_vertical_balance(self, vertical_balance: VerticalBalance):
        self.vertical_balance = vertical_balance

    def add_anchor(self, stage_id: int, anchor: Anchor, pre_tension: float) -> None:
        if not isinstance(self.anchors, Anchors):
            self.anchors = Anchors()
        if anchor.name not in self.anchors.anchor_names:
            self.anchors.anchors.append(anchor)

        self.construction_stages.stages[stage_id].anchors.append(
            AnchorOrStrutPresstressReference(
                name=anchor.name, pre_stress=pre_tension or _DEFAULT_PRE_STRESS
            )
        )

    def add_strut(self, stage_id: int, strut: Strut, pre_compression: float) -> None:
        if not isinstance(self.struts, Struts):
            self.struts = Struts()
        if strut.name not in self.struts.strut_names:
            self.struts.struts.append(strut)

        self.construction_stages.stages[stage_id].struts.append(
            AnchorOrStrutPresstressReference(
                name=strut.name, pre_stress=pre_compression or _DEFAULT_PRE_STRESS
            )
        )

    def add_load(
        self, load: Union[HorizontalLineLoad, Moment, NormalForce], stage_id: int
    ):
        if isinstance(load, HorizontalLineLoad):
            self.is_valid_unique_load_names(
                load_list=self.construction_stages.stages[stage_id].horizontal_line_loads,
                name=load.name,
            )
            if (
                isinstance(self.horizontal_line_loads, str)
                or self.horizontal_line_loads is None
            ):
                self.horizontal_line_loads = HorizontalLineLoads()
            if load not in self.horizontal_line_loads.loads:
                self.horizontal_line_loads.loads.append(load)
            self.construction_stages.stages[stage_id].horizontal_line_loads.append(
                load.name
            )
        elif isinstance(load, Moment):
            self.is_valid_unique_load_names(
                load_list=self.construction_stages.stages[stage_id].moment_loads,
                name=load.name,
            )
            if isinstance(self.moments, str) or self.moments is None:
                self.moments = Moments()
            if load not in self.moments.loads:
                self.moments.loads.append(load)
            self.construction_stages.stages[stage_id].moment_loads.append(load.name)
        elif isinstance(load, NormalForce):
            self.is_valid_unique_load_names(
                load_list=self.construction_stages.stages[stage_id].normal_forces,
                name=load.name,
            )
            if isinstance(self.normal_forces, str) or self.normal_forces is None:
                self.normal_forces = NormalForces()
            if load not in self.normal_forces.loads:
                self.normal_forces.loads.append(load)
            self.construction_stages.stages[stage_id].normal_forces.append(load.name)
        elif isinstance(load, UniformLoad):
            self.add_uniform_load(load=load, stage_id=stage_id)
        else:
            raise ValueError("This type of load is not yet supported")

    def add_uniform_load(self, stage_id: int, load: UniformLoad) -> None:
        self.is_valid_unique_load_names(
            load_list=self.construction_stages.stages[stage_id].uniform_loads,
            name=load.name,
        )
        if not isinstance(self.uniform_loads, UniformLoads):
            self.uniform_loads = UniformLoads()
        if load not in self.uniform_loads.loads:
            self.uniform_loads.loads.append(load)
        self.construction_stages.stages[stage_id].uniform_loads.append(load.name)

    def add_surcharge_load(self, stage_id: int, load: SurchargeLoad, side: Side) -> None:
        if not isinstance(self.surcharge_loads, SurchargeLoads):
            self.surcharge_loads = SurchargeLoads()
        if load not in self.surcharge_loads.loads:
            self.surcharge_loads.loads.append(load)
        # Add load to stages.
        stage = self.construction_stages.stages[stage_id]
        if side == Side.BOTH:
            self.is_valid_unique_load_names(
                load_list=stage.surcharge_loads_left, name=load.name
            )
            stage.surcharge_loads_left.append(load.name)
            self.is_valid_unique_load_names(
                load_list=stage.surcharge_loads_right, name=load.name
            )
            stage.surcharge_loads_right.append(load.name)
        elif side == Side.LEFT:
            self.is_valid_unique_load_names(
                load_list=stage.surcharge_loads_left, name=load.name
            )
            stage.surcharge_loads_left.append(load.name)
        elif side == Side.RIGHT:
            self.is_valid_unique_load_names(
                load_list=stage.surcharge_loads_right, name=load.name
            )
            stage.surcharge_loads_right.append(load.name)
        else:
            raise ValueError(f"Provide a Side, received {side}")

    def is_valid_unique_load_names(self, load_list: List[str], name: str) -> bool:
        """Load list should have unique names in list of loads."""
        if name in load_list:
            raise ValueError(
                f"{name} load name is duplicated. Please change the name of the load."
            )
        return True

    def add_spring_support(self, stage_id: int, support: Support) -> None:
        if not isinstance(self.spring_supports, SupportContainer):
            self.spring_supports = SupportContainer()
        if support.name not in self.spring_supports.support_names:
            self.spring_supports.supports.append(support)

        self.construction_stages.stages[stage_id].spring_supports.append(support.name)

    def add_rigid_support(self, stage_id: int, support: Support) -> None:
        if not isinstance(self.rigid_supports, SupportContainer):
            self.rigid_supports = SupportContainer()
        if support.name not in self.rigid_supports.support_names:
            self.rigid_supports.supports.append(support)

        self.construction_stages.stages[stage_id].rigid_supports.append(support.name)


class Resume(DSheetpilingWithNumberOfRowsTable):
    resume: List[Dict[str, float]]


class BreukData(DSeriesInlineReversedProperties):
    percentage_mobilized_resistance_left: float
    percentage_mobilized_resistance_right: float
    effective_left: float
    effective_right: float
    water_pressure_left: float
    water_pressure_right: float
    max_passive_effect_resistance_left: float
    max_passive_effect_resistance_right: float
    max_passive_moment_left: float
    max_passive_moment_right: float
    max_mobilized_passive_moment_left: float
    max_mobilized_passive_moment_right: float
    vertical_force_left: float
    vertical_force_right: float
    level_of_single_support: float
    node_of_single_support: float
    moment_present: float
    active_side: float


class MomentsForcesDisplacements(DSeriesWrappedTableStructure):
    momentsforcesdisplacements: List[Dict[str, float]]


class AnchorData(DSheetpilingWithNumberOfRowsTable):
    anchordata: List[Dict[str, Union[float, str]]]


class Pressures(DSheetpilingWithNumberOfRowsTable):
    pressures: List[Dict[str, Union[float, str]]]


class SideOutput(DSeriesStructure):
    calculation_method: Optional[str] = None
    water_level: Optional[str] = None
    surface: Optional[str] = None
    soil_profile_for_single_pile_model: Optional[str] = None
    soil_profile_for_sheet_piling_model: Optional[str] = None
    pressures: Optional[Pressures] = None
    force_from_layer: Optional[str] = None
    lambdas: Optional[str] = None
    slide_plane: Optional[str] = None
    vertical_balance_per_layer: Optional[str] = None


class OutputConstructionStage(DSeriesRepeatedGroupedProperties):
    anchor_data: Optional[AnchorData] = None
    hload_data: Optional[str] = None
    breuk_data: Optional[BreukData] = None
    passive_side_data: Optional[str] = None
    soil_collapse_data: Optional[str] = None
    moments_forces_displacements: Optional[MomentsForcesDisplacements] = None
    side: Optional[List[SideOutput]] = None
    uniform_load_data: Optional[str] = None
    horizontal_line_load_data: Optional[str] = None
    surcharge_data: Optional[str] = None
    normal_force_data: Optional[str] = None
    moment_data: Optional[str] = None
    support_data: Optional[str] = None
    vertical_balance_data: Optional[str] = None

    @classmethod
    def get_list_field_names(cls) -> List[str]:
        return ["side"]


class DesignLengthInfo(DSeriesWrappedTableStructure):
    designlengthinfo: List[Dict[str, float]]


class DesignLengthCalculation(DSeriesWrappedTableStructure):
    designlengthcalculation: List[Dict[str, float]]


class DesignSheetpileLength(DSeriesStructure):
    design_length_info: Optional[DesignLengthInfo] = None
    design_length_calculation: Optional[DesignLengthCalculation] = None


class PointsOnSheetpile(DSheetpilingWithNumberOfRowsTable):
    pointsonsheetpile: List[Dict[str, float]]


class CurAnchorForceResults(DSheetpilingWithNumberOfRowsTable):
    curanchorforceresults: List[Dict[str, float]]


class BaseVerificationStructureProperties(DSeriesRepeatedGroupedProperties):
    points_on_sheetpile: Optional[List[PointsOnSheetpile]] = None
    construction_stage: Optional[List[OutputConstructionStage]] = None

    @classmethod
    def get_list_field_names(cls) -> List[str]:
        return ["points_on_sheetpile", "construction_stage"]


class DSheetPilingOutputStructure(DSeriesRepeatedGroupedProperties):
    @classmethod
    def get_list_field_names(cls) -> List[str]:
        return ["points_on_sheetpile", "construction_stage"]

    @classmethod
    def parse_text(cls, text):
        return super().parse_text(text)

    calculation_type: str

    # General data
    sheet_pile_elements: Optional[str] = None
    calculated_displacements: Optional[str] = None

    # Standard, Kranz and Reliability calculation
    points_on_sheetpile: Optional[List[PointsOnSheetpile]] = None
    construction_stage: Optional[List[OutputConstructionStage]] = None

    # Design Sheet Pile Length calculation
    design_sheetpile_length: Optional[DesignSheetpileLength] = None

    # Settlement by Vibration calculation
    settlement_by_vibration: Optional[str] = None

    # Verify calculation including Overall Stability calculation
    overall_partial_factor_set: Optional[str] = None
    factors_for_overall_stability: Optional[str] = None
    overall_stability_results: Optional[str] = None

    # Verify calculation according to CUR or EC7-NL with method B
    factors_for_verification: Optional[str] = None

    # Verify calculation according to CUR or EC7-NL
    verify_step_6____5_serviceability_limit_state: Optional[
        BaseVerificationStructureProperties
    ] = None
    verify_step_6____5_multiplied_by_factor: Optional[
        BaseVerificationStructureProperties
    ] = None
    verify_step_6____1_low_modulus_of_subgrade_reaction_and_high_passive_water_level: Optional[
        BaseVerificationStructureProperties
    ] = None
    verify_step_6____2_high_modulus_of_subgrade_reaction_and_high_passive_water_level: Optional[
        BaseVerificationStructureProperties
    ] = None
    verify_step_6____3_low_modulus_of_subgrade_reaction_and_low_passive_water_level: Optional[
        BaseVerificationStructureProperties
    ] = None
    verify_step_6____4_high_modulus_of_subgrade_reaction_and_low_passive_water_level: Optional[
        BaseVerificationStructureProperties
    ] = None
    cur_anchor_force_results: Optional[CurAnchorForceResults] = None

    # Verify calculation according to EC7-BE or EC7-General
    verify_deformation_serviceability_limit_state: Optional[
        BaseVerificationStructureProperties
    ] = None
    eurocode_1_set_1: Optional[BaseVerificationStructureProperties] = None
    eurocode_1_set_2: Optional[BaseVerificationStructureProperties] = None
    eurocode_2: Optional[BaseVerificationStructureProperties] = None
    eurocode_3: Optional[BaseVerificationStructureProperties] = None
    eurocode_belgium_set_1: Optional[BaseVerificationStructureProperties] = None
    eurocode_belgium_set_2: Optional[BaseVerificationStructureProperties] = None

    # Kranz calculation
    angles_kranz_calculation: Optional[str] = None
    kranz_calculation: Optional[str] = None
    kranz_diagram_results: Optional[str] = None

    # Resumes
    resume: Optional[Resume] = None
    anchors_and_struts_resume: Optional[str] = None
    supports_resume: Optional[str] = None
    maximum_anchor_force: Optional[str] = None
    maximum_anchor_force_be_set_1: Optional[str] = None
    maximum_summary_results: Optional[str] = None
    maximum_summary_results_be_set_1: Optional[str] = None
    warnings: Optional[str] = None
    errors: Optional[str] = None


class DSheetPilingStructure(DSeriesPilingParserStructure):
    input_data: DSheetPilingInputStructure = DSheetPilingInputStructure()
    output_data: Optional[DSheetPilingOutputStructure] = None

    @property
    def is_valid(self) -> bool:
        return self.input_data.is_valid


class DSheetPilingDumpStructure(DSeriesPilingDumpParserStructure):
    dumpfile: DSheetPilingStructure



"""
Profile and Layer classes which are used by both D-Foundations and DSheetPiling.
D-Foundations often requires more parameters, which are unused for DSheetPiling.

"""

from typing import List

from pydantic import Field, field_validator
from typing_extensions import Annotated

from geolib.geometry.one import Point
from geolib.models import BaseDataClass

from .internal import SoilProfile as InternalSoilProfile


class CPT(BaseDataClass):
    """Internal Placeholder CPT measurements as used in D-Foundations/D-Sheetpiling.

    It is here because input files can contain CPTs coupled to a Profile.

    .. todo::
        * Add Friction and other parameters?
        * Unify with DFoundations
    """

    z: List[float]
    qc: List[float]


class CPTRule(BaseDataClass):
    """Placeholder class for CPT interpretation options."""


class SoilLayer(BaseDataClass):
    """Layer in a D-Sheetpiling Profile."""

    top_of_layer: float
    soil: str
    water_pressure_top: float = 0.0
    water_pressure_bottom: float = 0.0


class SoilProfile(BaseDataClass):
    """D-Sheetpiling Profile."""

    name: str
    layers: Annotated[List[SoilLayer], Field(min_length=1)]
    coordinate: Point = Point(x=0, y=0)

    @classmethod
    def top_of_layers_must_be_decreasing(cls, v):
        top_of_layers = [l.top_of_layer for l in v]
        if top_of_layers != sorted(top_of_layers, reverse=True):
            raise ValueError(
                f"Top of layer must decrease with each layer along depth. Top of layers: {top_of_layers}"
            )
        return v

    top_of_layer_validator = field_validator(
        "layers",
    )(top_of_layers_must_be_decreasing)

    def to_internal(self) -> InternalSoilProfile:
        return InternalSoilProfile(**self.model_dump(exclude_none=True))



import abc
from enum import Enum
from pathlib import Path
from typing import BinaryIO, List, Optional, Set, Type, Union

import matplotlib.pyplot as plt
from pydantic import DirectoryPath, FilePath
from shapely.geometry import LineString, Polygon
from shapely.ops import polygonize
from shapely.validation import make_valid

from geolib.geometry import Point
from geolib.models import BaseModel
from geolib.soils import Soil

from .analysis import DStabilityAnalysisMethod
from .dstability_parserprovider import DStabilityParserProvider
from .internal import (
    AnalysisType,
    BishopSlipCircleResult,
    CalculationSettings,
    DStabilityResult,
    DStabilityStructure,
    PersistableExcavation,
    PersistableLayer,
    PersistablePoint,
    PersistableSoil,
    PersistableStateCorrelation,
    Scenario,
    SoilCollection,
    SoilCorrelation,
    SoilLayerCollection,
    SoilVisualisation,
    SpencerSlipPlaneResult,
    UpliftVanSlipCircleResult,
    Waternet,
)
from .loads import Consolidation, DStabilityLoad
from .reinforcements import DStabilityReinforcement
from .serializer import DStabilityInputSerializer, DStabilityInputZipSerializer
from .states import DStabilityStateLinePoint, DStabilityStatePoint


class DStabilityCalculationType(Enum):
    """Set Type of Calculation."""

    BoundarySearch = 1
    SingleCalc = 2


class DStabilityCalculationModel(Enum):
    """Set Type of Calculation."""

    Bishop = 1
    UpliftVan = 2
    Spencer = 3


class DStabilityObject(BaseModel, metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def _to_dstability_sub_structure(self):
        raise NotImplementedError


class DStabilityModel(BaseModel):
    """D-Stability is software for soft soil slope stability.

    This model can read, modify and create
    .stix files
    """

    current_scenario: int = -1
    current_stage: int = -1
    current_calculation: int = -1
    datastructure: DStabilityStructure = DStabilityStructure()
    current_id: int = -1

    def __init__(self, *args, **data) -> None:
        super().__init__(*args, **data)
        self.current_id = self.datastructure.get_unique_id()

    @property
    def parser_provider_type(self) -> Type[DStabilityParserProvider]:
        return DStabilityParserProvider

    @property
    def default_console_path(self) -> Path:
        return Path("DStabilityConsole/D-Stability Console.exe")

    @property
    def custom_console_path(self) -> Path:
        return self.get_meta_property("dstability_console_path")

    @property
    def soils(self) -> SoilCollection:
        """Enables easy access to the soil in the internal dict-like datastructure. Also enables edit/delete for individual soils."""
        return self.datastructure.soils

    @property
    def soil_correlations(self) -> SoilCorrelation:
        return self.datastructure.soilcorrelation

    def _get_next_id(self) -> int:
        self.current_id += 1
        return self.current_id

    def parse(self, *args, **kwargs):
        super().parse(*args, **kwargs)
        self.current_id = self.datastructure.get_unique_id()

    @property
    def waternets(self) -> List[Waternet]:
        return self.datastructure.waternets

    @property
    def output(self) -> List[DStabilityResult]:
        def _get_result_or_none(scenario_index, calculation_index) -> DStabilityResult:
            if self.has_result(
                scenario_index=int(scenario_index),
                calculation_index=int(calculation_index),
            ):
                return self.get_result(
                    scenario_index=int(scenario_index),
                    calculation_index=int(calculation_index),
                )
            else:
                return None

        all_results = []

        for scenario_index, scenario in enumerate(self.datastructure.scenarios):
            for calculation_index, _ in enumerate(scenario.Calculations):
                all_results.append(
                    _get_result_or_none(
                        scenario_index=scenario_index, calculation_index=calculation_index
                    )
                )

        return all_results

    def has_result(
        self,
        scenario_index: Optional[int] = None,
        calculation_index: Optional[int] = None,
    ) -> bool:
        """
        Returns whether a calculation has a result.

        Args:
            scenario_index (Optional[int]): Index of a scenario, if None the current scenario is used.
            calculation_index (Optional[int]): Index of a calculation, if None the current calculation is used.

        Returns:
            bool: Value indicating whether the calculation has a result.
        """
        if calculation_index is None:
            calculation_index = self.current_calculation
        if scenario_index is None:
            scenario_index = self.current_scenario

        return self.datastructure.has_result(scenario_index, calculation_index)

    def get_result(
        self,
        scenario_index: Optional[int] = None,
        calculation_index: Optional[int] = None,
    ) -> DStabilityResult:
        """
        Returns the results of a calculation. Calculation results are based on analysis type and calculation type.

        Args:
            scenario_index (Optional[int]): Index of a scenario, if None is supplied the result of the current scenario is returned.
            calculation_index (Optional[int]): Index of a calculation, if None is supplied the result of the current calculation is returned.

        Returns:
            DStabilityResult: The analysis results of the stage.

        Raises:
            ValueError: No results or calculationsettings available
        """
        if calculation_index is None:
            calculation_index = self.current_calculation
        if scenario_index is None:
            scenario_index = self.current_scenario

        result = self._get_result_substructure(scenario_index, calculation_index)
        return result

    def _get_result_substructure(
        self, scenario_index: Optional[int], calculation_index: Optional[int]
    ) -> DStabilityResult:
        scenario_index = self.get_scenario_index(scenario_index)
        calculation_index = self.get_calculation_index(calculation_index)

        if self.datastructure.has_result(scenario_index, calculation_index):
            result_id = (
                self.datastructure.scenarios[scenario_index]
                .Calculations[calculation_index]
                .ResultId
            )
            calculation_settings = self._get_calculation_settings(
                scenario_index, calculation_index
            )
            analysis_type = calculation_settings.AnalysisType
            calculation_type = calculation_settings.CalculationType

            results = self.datastructure.get_result_substructure(
                analysis_type, calculation_type
            )

            for result in results:
                if result.Id == result_id:
                    return result

        raise ValueError(f"No result found for result id {calculation_index}")

    def _get_calculation_settings(
        self, scenario_index: int, calculation_index: int
    ) -> CalculationSettings:
        calculation_settings_id = (
            self.datastructure.scenarios[scenario_index]
            .Calculations[calculation_index]
            .CalculationSettingsId
        )

        for calculation_settings in self.datastructure.calculationsettings:
            if calculation_settings.Id == calculation_settings_id:
                return calculation_settings

        raise ValueError(
            f"No calculation settings found for calculation {calculation_index} in scenario {scenario_index}."
        )

    def get_slipcircle_result(
        self,
        scenario_index: Optional[int] = None,
        calculation_index: Optional[int] = None,
    ) -> Union[BishopSlipCircleResult, UpliftVanSlipCircleResult]:
        """
        Get the slipcircle(s) of the calculation result of a given stage.

        Args:
            scenario_index (Optional[int]): scenario for which to get the available results
            calculation_index (Optional[int]): calculation for which to get the available results

        Returns:
            Union[BishopSlipCircleResult, UpliftVanSlipCircleResult]: the slipcircle for the given calculation

        Raises:
            ValueError: Result is not available for provided scenario and calculation index
            AttributeError: When the result has no slipcircle. Try get the slipplane
        """
        result = self._get_result_substructure(scenario_index, calculation_index)
        return result.get_slipcircle_output()

    def get_slipplane_result(
        self,
        scenario_index: Optional[int] = None,
        calculation_index: Optional[int] = None,
    ) -> SpencerSlipPlaneResult:
        """
        Get the slipplanes of the calculations result of a calculation.

        Args:
            scenario_index (Optional[int]): scenario for which to get the available results
            calculation_index (Optional[int]): calculation for which to get the available results

        Returns:
            SpencerSlipPlaneResult: the slip plane for the given calculation

        Raises:
            ValueError: Result is not available for provided scenario and calculation index
            AttributeError: When the result has no slipplane. Try get the slipcircle
        """
        result = self._get_result_substructure(scenario_index, calculation_index)
        return result.get_slipplane_output()

    def _get_geometry(self, scenario_index: int, stage_index: int):
        geometry_id = (
            self.datastructure.scenarios[scenario_index].Stages[stage_index].GeometryId
        )

        for geometry in self.datastructure.geometries:
            if geometry.Id == geometry_id:
                return geometry

        raise ValueError(
            f"No geometry found for stage {stage_index} in scenario {scenario_index}."
        )

    def _get_soil_layers(self, scenario_index: int, stage_index: int):
        return self.datastructure._get_soil_layers(scenario_index, stage_index)

    def _get_waternet(self, scenario_index: int, stage_index: int):
        waternet_id = (
            self.datastructure.scenarios[scenario_index].Stages[stage_index].WaternetId
        )

        for waternet in self.datastructure.waternets:
            if waternet.Id == waternet_id:
                return waternet

        raise ValueError(
            f"No waternet found for stage {stage_index} in scenario {scenario_index}."
        )

    def _get_state(self, scenario_index: int, stage_index: int):
        state_id = (
            self.datastructure.scenarios[scenario_index].Stages[stage_index].StateId
        )

        for state in self.datastructure.states:
            if state.Id == state_id:
                return state

        raise ValueError(
            f"No state found for stage {stage_index} in scenario {scenario_index}."
        )

    def _get_state_correlations(self, scenario_index: int, stage_index: int):
        state_correlations_id = (
            self.datastructure.scenarios[scenario_index]
            .Stages[stage_index]
            .StateCorrelationsId
        )

        for state_correlations in self.datastructure.statecorrelations:
            if state_correlations.Id == state_correlations_id:
                return state_correlations

        raise ValueError(
            f"No state correlations found for stage {stage_index} in scenario {scenario_index}."
        )

    def _get_excavations(self, scenario_index: int, stage_index: int):
        return self.datastructure._get_excavations(scenario_index, stage_index)

    def _get_loads(self, scenario_index: int, stage_index: int):
        return self.datastructure._get_loads(scenario_index, stage_index)

    def _get_reinforcements(self, scenario_index: int, stage_index: int):
        reinforcements_id = (
            self.datastructure.scenarios[scenario_index]
            .Stages[stage_index]
            .ReinforcementsId
        )

        for reinforcements in self.datastructure.reinforcements:
            if reinforcements.Id == reinforcements_id:
                return reinforcements

        raise ValueError(
            f"No reinforcements found for stage {stage_index} in scenario {scenario_index}."
        )

    def serialize(self, location: Union[FilePath, DirectoryPath, BinaryIO]):
        """Support serializing to directory while developing for debugging purposes."""
        if isinstance(location, Path) and location.is_dir():
            serializer = DStabilityInputSerializer(ds=self.datastructure)
        else:
            serializer = DStabilityInputZipSerializer(ds=self.datastructure)
        serializer.write(location)
        if isinstance(location, Path):
            self.filename = location

    def add_scenario(
        self, label: str = "Scenario", notes: str = "", set_current: bool = True
    ) -> int:
        """Add a new scenario to the model.

        Args:
            label (str): Label for the scenario.
            notes (str): Notes for the scenario.
            set_current (bool): Whether to make the new scenario the current scenario.

        Returns:
            the id of the new stage
        """
        new_id = self._get_next_id()
        new_scenario_id, new_unique_id = self.datastructure.add_default_scenario(
            label, notes, new_id
        )

        if set_current:
            self.current_scenario = new_scenario_id
            self.current_stage = 0
            self.current_calculation = 0

        self.current_id = new_unique_id
        return new_scenario_id

    def add_stage(
        self,
        scenario_index: Optional[int] = None,
        label: str = "Stage",
        notes: str = "",
        set_current=True,
    ) -> int:
        """Add a new stage to the model at the given scenario index.

        Args:
            scenario_index (Optional[int]): The scenario index to add the stage to, defaults to the current scenario.
            label (str): Label for the stage.
            notes (str): Notes for the stage.
            set_current (bool): Whether to make the new stage the current stage.

        Returns:
            the id of the new stage
        """
        scenario_index = self.get_scenario_index(scenario_index)

        new_id = self._get_next_id()
        new_stage_index, new_unique_id = self.datastructure.add_default_stage(
            scenario_index, label, notes, new_id
        )

        if set_current:
            self.current_stage = new_stage_index
        self.current_id = new_unique_id
        return new_stage_index

    def add_calculation(
        self,
        scenario_index: Optional[int] = None,
        label: str = "Calculation",
        notes: str = "",
        set_current: bool = True,
    ) -> int:
        """Add a new calculation to the model.

        Args:
            scenario_index (Optional[int]): The scenario index to add the calculation to, defaults to the current scenario.
            label (str): Label for the calculation.
            notes (str): Notes for the calculation.
            set_current (bool): Whether to make the new calculation the current calculation.

        Returns:
            the id of the new stage
        """
        scenario_index = self.get_scenario_index(scenario_index)

        new_id = self._get_next_id()
        new_calculation_index, new_unique_id = self.datastructure.add_default_calculation(
            scenario_index, label, notes, new_id
        )

        if set_current:
            self.current_calculation = new_calculation_index
        self.current_id = new_unique_id
        return new_calculation_index

    @property
    def scenarios(self) -> List[Scenario]:
        return self.datastructure.scenarios

    def add_soil(self, soil: Soil) -> int:
        """
        Add a new soil to the model. The code must be unique, the id will be generated

        Args:
            soil (Soil): a new soil

        Returns:
            int: id of the added soil
        """
        if soil.code == None:
            raise ValueError("Soil.code may not be None")
        if self.soils.has_soil_code(soil.code):
            raise ValueError(f"The soil with code {soil.code} is already defined.")

        soil.id = str(self._get_next_id())
        dstability_soil = self.soils.add_soil(soil)
        return int(dstability_soil.Id)

    @property
    def points(self):
        """Enables easy access to the points in the internal dict-like datastructure. Also enables edit/delete for individual points."""

    def add_layer(
        self,
        points: List[Point],
        soil_code: str,
        label: str = "",
        notes: str = "",
        scenario_index: Optional[int] = None,
        stage_index: Optional[int] = None,
    ) -> int:
        """
        Add a soil layer to the model

        Args:
            points (List[Point]): list of Point classes, in clockwise order (non closed simple polygon)
            soil_code (str): code of the soil for this layer
            label (str): label defaults to empty string
            notes (str): notes defaults to empty string
            scenario_index (Optional[int]): scenario to add to, defaults to the current scenario
            stage_index (Optional[int]): stage to add to, defaults to the current stage

        Returns:
            int: id of the added layer
        """
        scenario_index = self.get_scenario_index(scenario_index)
        stage_index = self.get_stage_index(stage_index)

        geometry = self._get_geometry(scenario_index, stage_index)
        soil_layers = self._get_soil_layers(scenario_index, stage_index)

        # Check if we have the soil code
        if not self.soils.has_soil_code(soil_code):
            raise ValueError(
                f"The soil with code {soil_code} is not defined in the soil collection."
            )

        # Make sure the points are valid
        persistable_points = self.make_points_valid(points)

        # Create the new layer
        new_layer = PersistableLayer(
            Id=str(self._get_next_id()),
            Label=label,
            Points=persistable_points,
            Notes=notes,
        )

        # Add the layer to the geometry
        self.add_layer_and_connect_points(geometry.Layers, new_layer)

        # Add the connection between the layer and the soil to soillayers
        soil = self.soils.get_soil(soil_code)
        soil_layers.add_soillayer(layer_id=new_layer.Id, soil_id=soil.Id)
        return int(new_layer.Id)

    def make_points_valid(self, points: List[Point]) -> List[PersistablePoint]:
        valid_points = make_valid(self.geolib_points_to_shapely_polygon(points))
        return self.to_dstability_points(valid_points)

    def connect_layers(self, layer1: PersistableLayer, layer2: PersistableLayer):
        """Connects two polygons by adding a the missing points on the polygon edges. Returns the two new polygons."""
        linestring1 = self.to_shapely_linestring(layer1.Points)
        linestring2 = self.to_shapely_linestring(layer2.Points)

        # Create a union of the two polygons and polygonize it creating two connected polygons
        union = linestring1.union(linestring2)
        result = [geom for geom in polygonize(union)]

        # If the result has two polygons, we return them, otherwise we return the original polygons
        if len(result) == 2:
            return result[0].exterior, result[1].exterior
        else:
            return linestring1, linestring2

    def add_layer_and_connect_points(
        self, current_layers: List[PersistableLayer], new_layer: PersistableLayer
    ):
        """Adds a new layer to the list of layers and connects the points of the new layer to the existing layers."""

        current_layers.append(new_layer)

        # Check if the new layer intersects with any of the existing layers
        for layer in current_layers:
            if layer != new_layer and self.dstability_points_to_shapely_polygon(
                layer.Points
            ).exterior.intersects(
                self.dstability_points_to_shapely_polygon(new_layer.Points).exterior
            ):
                # If it does, connect the layers
                linestring1, linestring2 = self.connect_layers(layer, new_layer)

                # Update the points of the layers
                current_layers[current_layers.index(layer)].Points = (
                    self.to_dstability_points(linestring1)
                )
                current_layers[current_layers.index(new_layer)].Points = (
                    self.to_dstability_points(linestring2)
                )

    def to_shapely_linestring(self, points: List[PersistablePoint]) -> LineString:
        converted_points = [(p.X, p.Z) for p in points]
        converted_points.append(converted_points[0])
        return LineString(converted_points)

    def dstability_points_to_shapely_polygon(
        self, points: List[PersistablePoint]
    ) -> Polygon:
        return Polygon([(p.X, p.Z) for p in points])

    def geolib_points_to_shapely_polygon(self, points: List[Point]) -> Polygon:
        return Polygon([(p.x, p.z) for p in points])

    def to_dstability_points(
        self, shapely_object: Union[LineString, Polygon]
    ) -> List[PersistablePoint]:
        if isinstance(shapely_object, LineString):
            coords = shapely_object.coords
        elif isinstance(shapely_object, Polygon):
            coords = shapely_object.exterior.coords
        else:
            raise ValueError(
                "shapely_object must be a LineString or Polygon, not {}".format(
                    type(shapely_object)
                )
            )

        persistable_points = [PersistablePoint(X=p[0], Z=p[1]) for p in list(coords)]

        # Remove duplicate points
        persistable_points = [
            i
            for n, i in enumerate(persistable_points)
            if i not in persistable_points[n + 1 :]
        ]

        # Remove last point if it is the same as the first
        if persistable_points[0] == persistable_points[-1]:
            persistable_points.pop(-1)

        return persistable_points

    def get_soil(self, code: str) -> PersistableSoil:
        """
        Gets an existing soil with the given soil code.

        Args:
            code (str): the code of the soil

        Returns:
            PersistableSoil: the soil
        """
        return self.soils.get_soil(code=code)

    def get_soil_by_name(self, name: str) -> PersistableSoil:
        """
        Gets an existing soil with the given soil name.

        Args:
            name (str): the name of the soil

        Returns:
            PersistableSoil: the soil
        """
        return self.soils.get_soil_by_name(name=name)

    def add_head_line(
        self,
        points: List[Point],
        label: str = "",
        notes: str = "",
        is_phreatic_line: bool = False,
        scenario_index: Optional[int] = None,
        stage_index: Optional[int] = None,
    ) -> int:
        """
        Add head line to the model

        Args:
            points (List[Point]): list of Point classes
            label (str): label defaults to empty string
            notes (str): notes defaults to empty string
            is_phreatic_line (bool): set as phreatic line, defaults to False
            scenario_index (Optional[int]): scenario to add to, defaults to the current scenario
            stage_index (Optional[int]): stage to add to, defaults to the current stage

        Returns:
            bool: id of the added headline
        """
        scenario_index = self.get_scenario_index(scenario_index)
        stage_index = self.get_stage_index(stage_index)

        waternet = self._get_waternet(scenario_index, stage_index)

        persistable_headline = waternet.add_head_line(
            str(self._get_next_id()), label, notes, points, is_phreatic_line
        )
        return persistable_headline.Id

    def add_reference_line(
        self,
        points: List[Point],
        bottom_headline_id: Optional[str] = None,
        top_head_line_id: Optional[str] = None,
        label: str = "",
        notes: str = "",
        scenario_index: Optional[int] = None,
        stage_index: Optional[int] = None,
    ) -> int:
        """
        Add reference line to the model

        Args:
            points (List[Point]): list of Point classes
            bottom_headline_id (int): id of the headline to use as the bottom headline
            top_head_line_id (int): id of the headline to use as the top headline
            label (str): label defaults to empty string
            notes (str): notes defaults to empty string
            scenario_index (Optional[int]): scenario to add to, defaults to the current scenario
            stage_index (Optional[int]): stage to add to, defaults to the current stage

        Returns:
            int: id of the added reference line
        """
        scenario_index = self.get_scenario_index(scenario_index)
        stage_index = self.get_stage_index(stage_index)

        waternet = self._get_waternet(scenario_index, stage_index)

        persistable_reference_line = waternet.add_reference_line(
            str(self._get_next_id()),
            label,
            notes,
            points,
            bottom_headline_id,
            top_head_line_id,
        )
        return int(persistable_reference_line.Id)

    def add_state_point(
        self,
        state_point: DStabilityStatePoint,
        scenario_index: Optional[int] = None,
        stage_index: Optional[int] = None,
    ) -> int:
        """
        Add state point to the model

        Args:
            state_point (DStabilityStatePoint): DStabilityStatePoint class
            scenario_index (Optional[int]): scenario to add to, defaults to the current scenario
            stage_index (Optional[int]): stage to add to, defaults to the current stage

        Returns:
            int: id of the added add_state_point
        """
        scenario_index = self.get_scenario_index(scenario_index)
        stage_index = self.get_stage_index(stage_index)

        states = self._get_state(scenario_index, stage_index)

        try:
            _ = self._get_geometry(scenario_index, stage_index).get_layer(
                state_point.layer_id
            )
        except ValueError:
            raise ValueError(f"No layer with id '{state_point.layer_id} in this geometry")

        state_point.id = (
            self._get_next_id()
        )  # the user does not know the id so we have to add it
        persistable_state_point = state_point._to_internal_datastructure()
        states.add_state_point(persistable_state_point)
        return int(persistable_state_point.Id)

    def add_state_line(
        self,
        points: List[Point],
        state_points: List[DStabilityStateLinePoint],
        scenario_index: Optional[int] = None,
        stage_index: Optional[int] = None,
    ) -> int:
        """
        Add state line. From the Soils, only the state parameters are used.

        points are a list of points with x,z coordinates
        state_point are a list of DStabilityStateLinePoint where ONLY the x is used, the Z will be calculated

        Args:
            points (List[Point]): The geometry points of the state line.
            state_point (List[DStabilityStatePoint]): The list of state point values.
            scenario_index (Optional[int]): scenario to add to, defaults to the current scenario.
            stage_index (Optional[int]): stage to add to, defaults to the current stage.

        Returns:
            PersistableStateLine: The created state line
        """
        scenario_index = self.get_scenario_index(scenario_index)
        stage_index = self.get_stage_index(stage_index)

        states = self._get_state(scenario_index, stage_index)

        # each point should belong to a layer
        persistable_points = []

        for point in points:
            point.id = self._get_next_id()  # assign a new id
            persistable_points.append(PersistablePoint(X=point.x, Z=point.z))

        persistable_state_line_points = []
        for state_point in state_points:
            state_point.id = self._get_next_id()  # assign a new id
            persistable_state_line_points.append(state_point._to_internal_datastructure())

        return states.add_state_line(persistable_points, persistable_state_line_points)

    def add_state_correlation(
        self,
        correlated_state_ids: List[int],
        scenario_index: Optional[int] = None,
        stage_index: Optional[int] = None,
    ):
        """
        Add state correlation between the given state point ids.

        Args:
            correlated_state_ids (List[int]): The state point ids to correlate.
            scenario_index (Optional[int]): scenario to add to, defaults to the current scenario.
            stage_index (Optional[int]): stage to add to, defaults to the current stage.

        Returns:
            None
        """
        scenario_index = self.get_scenario_index(scenario_index)
        stage_index = self.get_stage_index(stage_index)

        state_correlations = self._get_state_correlations(scenario_index, stage_index)

        for state_id in correlated_state_ids:
            try:
                _ = self._get_state(scenario_index, stage_index).get_state(state_id)
            except ValueError:
                raise ValueError(f"No state point with id '{state_id} in this geometry")

        persistable_state_correlation = PersistableStateCorrelation(
            CorrelatedStateIds=correlated_state_ids, IsFullyCorrelated=True
        )

        state_correlations.add_state_correlation(persistable_state_correlation)

    def add_excavation(
        self,
        points: List[Point],
        label: str,
        notes: str = "",
        scenario_index: Optional[int] = None,
        stage_index: Optional[int] = None,
    ):
        scenario_index = self.get_scenario_index(scenario_index)
        stage_index = self.get_stage_index(stage_index)

        persistable_excavation = PersistableExcavation(
            Label=label,
            Notes=notes,
            Points=[PersistablePoint(X=p.x, Z=p.z) for p in points],
        )
        self._get_excavations(scenario_index, stage_index).append(persistable_excavation)

    def add_load(
        self,
        load: DStabilityLoad,
        consolidations: Optional[List[Consolidation]] = None,
        scenario_index: Optional[int] = None,
        stage_index: Optional[int] = None,
    ) -> None:
        """Add a load to the object.

        The geometry should be defined before adding loads.

        If no consolidations are provided, a Consolidation with default values will be made for each SoilLayer.
        It is not possible to set consolidation degrees of loads afterwards since they don't have an id.

        Args:
            load: A subclass of DStabilityLoad.
            scenario_index (Optional[int]): scenario to add to, defaults to the current scenario.
            stage_index (Optional[int]): stage to add to, defaults to the current stage.

        Raises:
            ValueError: When the provided load is no subclass of DStabilityLoad, an invalid stage_index is provided, or the datastructure is no longer valid.
        """
        scenario_index = self.get_scenario_index(scenario_index)
        stage_index = self.get_stage_index(stage_index)

        if not issubclass(type(load), DStabilityLoad):
            raise ValueError(
                f"load should be a subclass of DstabilityReinforcement, received {load}"
            )
        if self.datastructure.has_soil_layers(
            scenario_index, stage_index
        ) and self.datastructure.has_loads(scenario_index, stage_index):
            if consolidations is None:
                consolidations = self._get_default_consolidations(
                    scenario_index, stage_index
                )
            else:
                self._verify_consolidations(consolidations, scenario_index, stage_index)
            self._get_loads(scenario_index, stage_index).add_load(load, consolidations)
        else:
            raise ValueError(
                f"No loads found for scenario {scenario_index} stage {stage_index}"
            )

    def add_soil_layer_consolidations(
        self,
        soil_layer_id: int,
        consolidations: Optional[List[Consolidation]] = None,
        scenario_index: Optional[int] = None,
        stage_index: Optional[int] = None,
    ) -> None:
        """Add consolidations for a layer (layerload).

        Consolidations cannot be added when adding soil layers since in the consolidations, all other soil layers need to be referred.
        Therefore, all soillayers in a stage should be defined before setting consolidation and
        the number of consolidations given should equal the amount of layers.

        Args:
            soil_layer_id: Consolidation is set for this soil layer id.
            consolidations: List of Consolidation. Must contain a Consolidation for every other layer.
            scenario_index (Optional[int]): scenario to add to, defaults to the current scenario.
            stage_index (Optional[int]): stage to add to, defaults to the current stage.

        Raises:
            ValueError: When the provided load is no subclass of DStabilityLoad, an invalid stage_index is provided, or the datastructure is no longer valid.
        """
        scenario_index = self.get_scenario_index(scenario_index)
        stage_index = self.get_stage_index(stage_index)

        if self.datastructure.has_soil_layer(
            scenario_index, stage_index, soil_layer_id
        ) and self.datastructure.has_loads(scenario_index, stage_index):
            if consolidations is None:
                consolidations = self._get_default_consolidations(
                    scenario_index, stage_index, soil_layer_id
                )
            else:
                self._verify_consolidations(
                    consolidations, scenario_index, stage_index, soil_layer_id
                )

            self._get_loads(scenario_index, stage_index).add_layer_load(
                soil_layer_id, consolidations
            )
        else:
            raise ValueError(
                f"No soil layer loads found for scenario {scenario_index} stage {stage_index}"
            )

    def _get_default_consolidations(
        self,
        scenario_index: int,
        stage_index: int,
        exclude_soil_layer_id: Optional[int] = None,
    ) -> List[Consolidation]:
        """Length of the consolidations is equal to the amount of soil layers.

        If exclude_soil_layer_id is provided, that specific soil layer id is not included in the consolidations.
        """
        if self.datastructure.has_soil_layers(scenario_index, stage_index):
            soil_layer_ids = self._get_soil_layers(scenario_index, stage_index).get_ids(
                exclude_soil_layer_id
            )
            return [Consolidation(layer_id=layer_id) for layer_id in soil_layer_ids]

        raise ValueError(f"No soil layers found for stage at index {stage_index}")

    def _verify_consolidations(
        self,
        consolidations: List[Consolidation],
        scenario_index: int,
        stage_index: int,
        exclude_soil_layer_id: Optional[int] = None,
    ) -> None:
        if self.datastructure.has_soil_layers(scenario_index, stage_index):
            consolidation_soil_layer_ids: Set[str] = {
                str(c.layer_id) for c in consolidations
            }
            soil_layer_ids = self._get_soil_layers(scenario_index, stage_index).get_ids(
                exclude_soil_layer_id
            )

            if consolidation_soil_layer_ids != soil_layer_ids:
                raise ValueError(
                    f"Received consolidations ({consolidation_soil_layer_ids}) should contain all soil layer ids ({soil_layer_ids})"
                )
        else:
            raise ValueError(f"No soil layers found for stage at index {stage_index}")

    def add_reinforcement(
        self,
        reinforcement: DStabilityReinforcement,
        scenario_index: Optional[int] = None,
        stage_index: Optional[int] = None,
    ) -> None:
        """Add a reinforcement to the model.

        Args:
            reinforcement: A subclass of DStabilityReinforcement.
            scenario_index (Optional[int]): scenario to add to, defaults to the current scenario.
            stage_index (Optional[int]): stage to add to, defaults to the current stage.

        Returns:
            int: Assigned id of the reinforcements (collection object of all reinforcements for a stage).

        Raises:
            ValueError: When the provided reinforcement is no subclass of DStabilityReinforcement, an invalid stage_index is provided, or the datastructure is no longer valid.
        """
        scenario_index = self.get_scenario_index(scenario_index)
        stage_index = self.get_stage_index(stage_index)

        if not issubclass(type(reinforcement), DStabilityReinforcement):
            raise ValueError(
                f"reinforcement should be a subclass of DstabilityReinforcement, received {reinforcement}"
            )

        if self.datastructure.has_reinforcements(scenario_index, stage_index):
            self._get_reinforcements(scenario_index, stage_index).add_reinforcement(
                reinforcement
            )
        else:
            raise ValueError(
                f"No reinforcements found for scenario {scenario_index} stage {stage_index}"
            )

    def add_soil_correlation(self, list_correlated_soil_ids: List[str]):
        """Add a soil correlation to the model.

        Args:
            list_correlated_soil_ids: A list of soil ids that are correlated.
        """
        self.soil_correlations.add_soil_correlation(list_correlated_soil_ids)

    def set_model(
        self,
        analysis_method: DStabilityAnalysisMethod,
        scenario_index: Optional[int] = None,
        calculation_index: Optional[int] = None,
    ) -> None:
        """Sets the model and applies the given parameters

        Args:
            analysis_method (DStabilityAnalysisMethod): A subclass of DStabilityAnalysisMethod.
            scenario_index (Optional[int]): scenario to add to, defaults to the current scenario
            calculation_index (Optional[int]): calculation to add to, defaults to the current calculation

        Raises:
            ValueError: When the provided analysis method is no subclass of DStabilityAnalysisMethod,
            an invalid stage_index is provided, the analysis method is not known or the datastructure is no longer valid.
        """
        scenario_index = self.get_scenario_index(scenario_index)
        calculation_index = self.get_calculation_index(calculation_index)

        calculationsettings = self._get_calculation_settings(
            scenario_index, calculation_index
        )

        _analysis_method_mapping = {
            AnalysisType.BISHOP: calculationsettings.set_bishop,
            AnalysisType.BISHOP_BRUTE_FORCE: calculationsettings.set_bishop_brute_force,
            AnalysisType.SPENCER: calculationsettings.set_spencer,
            AnalysisType.SPENCER_GENETIC: calculationsettings.set_spencer_genetic,
            AnalysisType.UPLIFT_VAN: calculationsettings.set_uplift_van,
            AnalysisType.UPLIFT_VAN_PARTICLE_SWARM: calculationsettings.set_uplift_van_particle_swarm,
        }

        try:
            _analysis_method_mapping[analysis_method.analysis_type](
                analysis_method._to_internal_datastructure()
            )
        except KeyError:
            raise ValueError(
                f"Unknown analysis method {analysis_method.analysis_type.value} found"
            )

    def get_scenario_index(self, scenario_index: Optional[int]):
        if scenario_index is None:
            return self.current_scenario
        else:
            return scenario_index

    def get_stage_index(self, stage_index: Optional[int]):
        if stage_index is None:
            return self.current_stage
        else:
            return stage_index

    def get_calculation_index(self, calculation_index: Optional[int]):
        if calculation_index is None:
            return self.current_calculation
        else:
            return calculation_index

    @staticmethod
    def get_soil_id_from_layer_id(
        layers: SoilLayerCollection, layer_id: str
    ) -> Union[str, None]:
        for layer in layers.SoilLayers:
            if layer.LayerId == layer_id:
                return layer.SoilId
        return None

    @staticmethod
    def get_color_from_soil_id(
        soil_visualizations: SoilVisualisation, soil_id: str
    ) -> str:
        for soil_visualization in soil_visualizations.SoilVisualizations:
            if soil_visualization.SoilId == soil_id:
                return soil_visualization.Color
        return "#000000"

    def _get_color_of_layer(
        self, layers_collection: SoilLayerCollection, layer: PersistableLayer
    ) -> str:
        layer_id = layer.Id
        # use the layer id to get the soil type id
        soil_type_id = DStabilityModel.get_soil_id_from_layer_id(
            layers_collection, layer_id
        )
        # get the color of the soil type
        color = DStabilityModel.get_color_from_soil_id(
            self.input.soilvisualizations, soil_type_id
        )
        return color.replace("#80", "#")

    def plot(
        self, scenario_index: Optional[int] = None, stage_index: Optional[int] = None
    ):
        geometry = self._get_geometry(scenario_index, stage_index)
        layers_collection = self._get_soil_layers(scenario_index, stage_index)
        fig, ax = plt.subplots()
        # loop over the layers
        for layer in geometry.Layers:
            # get list of x and y coordinates
            x = [p.X for p in layer.Points]
            y = [p.Z for p in layer.Points]
            # get color of layer
            color = self._get_color_of_layer(layers_collection, layer)
            # create a polygon
            ax.fill(x, y, color=color)
        plt.axis("off")
        return fig, ax



import abc
from typing import List

from pydantic import Field, PositiveInt
from typing_extensions import Annotated

from geolib.models import BaseDataClass

from ...geometry.one import Point
from ...utils import snake_to_camel
from .internal import (
    AnalysisType,
    NullablePersistablePoint,
    OptionsType,
    PersistableBishopBruteForceSettings,
    PersistableBishopSettings,
    PersistableCircle,
    PersistableGeneticSlipPlaneConstraints,
    PersistableGridEnhancements,
    PersistablePoint,
    PersistableSearchArea,
    PersistableSearchGrid,
    PersistableSlipPlaneConstraints,
    PersistableSpencerGeneticSettings,
    PersistableSpencerSettings,
    PersistableTangentArea,
    PersistableTangentLines,
    PersistableTwoCirclesOnTangentLine,
    PersistableUpliftVanParticleSwarmSettings,
    PersistableUpliftVanSettings,
)


class DStabilityObject(BaseDataClass, metaclass=abc.ABCMeta):
    """Base Class for objects in the analysis module."""

    @abc.abstractmethod
    def _to_internal_datastructure(self):
        """Convert to internal datastructure"""
        raise NotImplementedError


class DStabilityAnalysisMethod(DStabilityObject):
    _analysis_type: AnalysisType

    @property
    def analysis_type(self) -> AnalysisType:
        return self._analysis_type


class DStabilityCircle(DStabilityObject):
    """Generates a circle object

    Args:
        center (Point): Center of the circle.
        radius (float): Radius of the circle.
    """

    center: Point
    radius: Annotated[float, Field(gt=0)]

    def _to_internal_datastructure(self) -> PersistableCircle:
        return PersistableCircle(
            Center=PersistablePoint(X=self.center.x, Z=self.center.z), Radius=self.radius
        )


class DStabilitySearchGrid(DStabilityObject):
    """Generates a search grid object

    Args:
        bottom_left (Point): Bottom left position of the search grid.
        number_of_points_in_x (int): Number of points to the right
        number_of_points_in_z (int): Number op points upwards
        space (float): Space between the points in x and z direction
    """

    bottom_left: Point
    number_of_points_in_x: PositiveInt
    number_of_points_in_z: PositiveInt
    space: Annotated[float, Field(gt=0)]

    def _to_internal_datastructure(self) -> PersistableSearchGrid:
        return PersistableSearchGrid(
            BottomLeft=NullablePersistablePoint(
                X=self.bottom_left.x, Z=self.bottom_left.z
            ),
            NumberOfPointsInX=self.number_of_points_in_x,
            NumberOfPointsInZ=self.number_of_points_in_z,
            Space=self.space,
        )


class DStabilitySearchArea(DStabilityObject):
    """Generates a search area object

    Args:
        height (float): Height of the search area.
        top_left (Point): Top left position of the search area
        width (float): Width of the search area
    """

    height: Annotated[float, Field(gt=0)]
    top_left: Point
    width: Annotated[float, Field(gt=0)]

    def _to_internal_datastructure(self) -> PersistableSearchArea:
        return PersistableSearchArea(
            Height=self.height,
            TopLeft=NullablePersistablePoint(X=self.top_left.x, Z=self.top_left.z),
            Width=self.width,
        )


class DStabilitySlipPlaneConstraints(DStabilityObject):
    """Generates a slip plane constraints object

    Args:
        is_size_constraints_enabled (bool): enabel size constraints, defaults to False
        is_zone_a_constraints_enabled (bool): enable constraints for zone A, defaults to False
        is_zone_b_constraints_enabled (bool): enable constraints for zone B, defaults to False
        minimum_slip_plane_depth (float): minimum slip plane depth, defaults to 0.0
        minimum_slip_plane_length (float): minimum slip plane length, defaults to 0.0
        width_zone_a (float): width of zone A, defaults to 0.0
        width_zone_b (float): width of zone B, defaults to 0.0
        x_left_zone_a (float): x coordinate of left point of zone A, defaults to 0.0
        x_left_zone_b (float): x coordinate of left point of zone B, defaults to 0.0
    """

    is_size_constraints_enabled: bool = False
    is_zone_a_constraints_enabled: bool = False
    is_zone_b_constraints_enabled: bool = False
    minimum_slip_plane_depth: float = 0.0
    minimum_slip_plane_length: float = 0.0
    width_zone_a: float = 0.0
    width_zone_b: float = 0.0
    x_left_zone_a: float = 0.0
    x_left_zone_b: float = 0.0

    def _to_internal_datastructure(self) -> PersistableSlipPlaneConstraints:
        data = {
            **{snake_to_camel(name): value for name, value in self.model_dump().items()}
        }
        return PersistableSlipPlaneConstraints(**data)


class DStabilityGeneticSlipPlaneConstraints(DStabilityObject):
    """Generates slip plane constraints object for the Spencer method

    Args:
        is_enabled (bool): Height of the search area.
        minimum_angle_between_slices (float): Minimum angle between slices, defaults to 0.
        minimum_thrust_line_percentage_inside_slices (float): Minimum thrustline percentage inside slices, defaults to 0.
    """

    is_enabled: bool = False
    minimum_angle_between_slices: float = 0.0
    minimum_thrust_line_percentage_inside_slices: float = 0.0

    def _to_internal_datastructure(self) -> PersistableGeneticSlipPlaneConstraints:
        data = {
            **{snake_to_camel(name): value for name, value in self.model_dump().items()}
        }
        return PersistableGeneticSlipPlaneConstraints(**data)


class DStabilityBishopAnalysisMethod(DStabilityAnalysisMethod):
    """Generates slip plane constraints object for the Bishop method

    Args:
        circle (DStabilityCircle): Bishop circle definition
    """

    _analysis_type: AnalysisType = AnalysisType.BISHOP
    circle: DStabilityCircle

    def _to_internal_datastructure(self) -> PersistableBishopSettings:
        return PersistableBishopSettings(Circle=self.circle._to_internal_datastructure())


class DStabilityBishopBruteForceAnalysisMethod(DStabilityAnalysisMethod):
    """Generates slip plane constraints object for the Bishop Brute Force method

    Args:
        extrapolate_search_space (bool): Extrapolate the search space, defaults to True
        search_grid (DStabilitySearchGrid):
        slip_plane_constraints (DStabilitySlipPlaneConstraints):
        bottom_tangent_line_z (float):
        number_of_tangent_lines (int):
        space_tangent_lines (float):
    """

    _analysis_type: AnalysisType = AnalysisType.BISHOP_BRUTE_FORCE
    extrapolate_search_space: bool = True
    search_grid: DStabilitySearchGrid
    slip_plane_constraints: DStabilitySlipPlaneConstraints = (
        DStabilitySlipPlaneConstraints()
    )
    bottom_tangent_line_z: float
    number_of_tangent_lines: PositiveInt
    space_tangent_lines: Annotated[float, Field(gt=0)]  # type: ignore

    def _to_internal_datastructure(self) -> PersistableBishopBruteForceSettings:
        return PersistableBishopBruteForceSettings(
            GridEnhancements=PersistableGridEnhancements(
                ExtrapolateSearchSpace=self.extrapolate_search_space
            ),
            SearchGrid=self.search_grid._to_internal_datastructure(),
            SlipPlaneConstraints=self.slip_plane_constraints._to_internal_datastructure(),
            TangentLines=PersistableTangentLines(
                BottomTangentLineZ=self.bottom_tangent_line_z,
                NumberOfTangentLines=self.number_of_tangent_lines,
                Space=self.space_tangent_lines,
            ),
        )


class DStabilitySpencerAnalysisMethod(DStabilityAnalysisMethod):
    """Generates slip plane for the Spencer method

    Args:
        slipplane ([Point]): The points of the slipplane
    """

    _analysis_type: AnalysisType = AnalysisType.SPENCER
    slipplane: List[Point]
    slip_plane_constraints: DStabilityGeneticSlipPlaneConstraints = (
        DStabilityGeneticSlipPlaneConstraints()
    )

    def _to_internal_datastructure(self) -> PersistableSpencerSettings:
        return PersistableSpencerSettings(
            SlipPlane=[
                PersistablePoint(X=point.x, Z=point.z) for point in self.slipplane
            ],
            SlipPlaneConstraints=self.slip_plane_constraints._to_internal_datastructure(),
        )


class DStabilitySpencerGeneticAnalysisMethod(DStabilityAnalysisMethod):
    """Generates the input for the genetic spencer algorithm

    Args:
        options_type (OptionsType): DEFAULT or THOROUGH, defaults to DEFAULT
        slip_plane_a (List[Point]): upper slip plane boundary
        slip_plane_b (List[Point]): lower slip line boundary
        slip_plane_constraints (DStabilityGeneticSlipPlaneConstraints): constraints for the slip plane
    """

    _analysis_type: AnalysisType = AnalysisType.SPENCER_GENETIC
    options_type: OptionsType = OptionsType.DEFAULT
    slip_plane_a: List[Point]
    slip_plane_b: List[Point]
    slip_plane_constraints: DStabilityGeneticSlipPlaneConstraints = (
        DStabilityGeneticSlipPlaneConstraints()
    )

    def _to_internal_datastructure(self) -> PersistableSpencerGeneticSettings:
        return PersistableSpencerGeneticSettings(
            OptionsType=self.options_type,
            SlipPlaneA=[
                PersistablePoint(X=point.x, Z=point.z) for point in self.slip_plane_a
            ],
            SlipPlaneB=[
                PersistablePoint(X=point.x, Z=point.z) for point in self.slip_plane_b
            ],
            SlipPlaneConstraints=self.slip_plane_constraints._to_internal_datastructure(),
        )


class DStabilityUpliftVanAnalysisMethod(DStabilityAnalysisMethod):
    """Generates the input for the uplift van analysis

    Args:
        first_circle (DStabilityCircle): The location of the left circle for the slipplane.
        second_circle_center (Point): The center of the circle on the right side of the slipplane
    """

    _analysis_type: AnalysisType = AnalysisType.UPLIFT_VAN
    first_circle: DStabilityCircle
    second_circle_center: Point

    def _to_internal_datastructure(self) -> PersistableUpliftVanSettings:
        return PersistableUpliftVanSettings(
            SlipPlane=PersistableTwoCirclesOnTangentLine(
                FirstCircleCenter=NullablePersistablePoint(
                    X=self.first_circle.center.x, Z=self.first_circle.center.z
                ),
                FirstCircleRadius=self.first_circle.radius,
                SecondCircleCenter=NullablePersistablePoint(
                    X=self.second_circle_center.x, Z=self.second_circle_center.z
                ),
            )
        )


class DStabilityUpliftVanParticleSwarmAnalysisMethod(DStabilityAnalysisMethod):
    """Generates the input for the uplift van analysis

    Args:
        options_type (OptionsType): DEFAULT or THOROUGH, defaults to DEFAULT
        search_area_a (DStabilitySearchArea): The search area for the circle on the left side of the slip plane
        search_area_b (DStabilitySearchArea): The search area for the circle on the right side of the slip plane
        slip_plane_constraints (DStabilitySlipPlaneConstraints): Slip plane constraints
        tangent_area_height (float): height of the tangent lines search area
        tangent_area_top_z (float): top z coordinate of the tangent lines area
    """

    _analysis_type: AnalysisType = AnalysisType.UPLIFT_VAN_PARTICLE_SWARM
    options_type: OptionsType = OptionsType.DEFAULT
    search_area_a: DStabilitySearchArea
    search_area_b: DStabilitySearchArea
    slip_plane_constraints: DStabilitySlipPlaneConstraints = (
        DStabilitySlipPlaneConstraints()
    )
    tangent_area_height: float
    tangent_area_top_z: float

    def _to_internal_datastructure(self) -> PersistableUpliftVanParticleSwarmSettings:
        return PersistableUpliftVanParticleSwarmSettings(
            OptionsType=self.options_type,
            SearchAreaA=self.search_area_a._to_internal_datastructure(),
            SearchAreaB=self.search_area_b._to_internal_datastructure(),
            SlipPlaneConstraints=self.slip_plane_constraints._to_internal_datastructure(),
            TangentArea=PersistableTangentArea(
                Height=self.tangent_area_height, TopZ=self.tangent_area_top_z
            ),
        )



"""
This module handles the three types of state types in DStability.
"""

import abc

from geolib.models import BaseDataClass

from ...geometry.one import Point
from ...utils import snake_to_camel
from .internal import (
    InternalStateTypeEnum,
    PersistablePoint,
    PersistableStateLinePoint,
    PersistableStatePoint,
    PersistableStochasticParameter,
    PersistableStress,
)


class DStabilityObject(BaseDataClass, metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def _to_internal_datastructure(self):
        raise NotImplementedError


class DStabilityStress(DStabilityObject):
    """DStability Stress

    Args:
        ocr (float): OCR value, defaults to 1.0
        pop (float): POP value, defaults to 0.0
        stochastic_parameter (PersistableStochasticParameter)
        state_type (InternalStateTypeEnum): type of state
    """

    ocr: float = 1.0
    pop: float = 0.0
    stochastic_parameter: PersistableStochasticParameter = (
        PersistableStochasticParameter()
    )
    state_type: InternalStateTypeEnum = InternalStateTypeEnum.POP

    def _to_internal_datastructure(self) -> PersistableStress:
        data = {
            **{snake_to_camel(name): value for name, value in self.model_dump().items()}
        }
        data["PopStochasticParameter"] = data.pop("StochasticParameter")
        return PersistableStress(**data)


class DStabilityStatePoint(DStabilityObject):
    """DStability StatePoint

    Args:
        id (int): id of the statepoint
        layer_id (int): id of the layer to add the statepoint, note that the API does not check if this point is within the layer
        pop (float): POP value, defaults to 0.0
        point (Point): location of the statepoint
        stress (DStabilityStress): DStabilityStress object
        is_probabilistic (bool): is probabilistic, default to false
        label (str): label of the statepoint

    """

    id: int = -1  # this will be filled in by the datamodel, not meant for the user
    layer_id: int
    point: Point
    stress: DStabilityStress
    is_probabilistic: bool = False
    label: str = ""

    def _to_internal_datastructure(self) -> PersistableStatePoint:
        model_dump = self.model_dump()
        data = {
            **{
                snake_to_camel(name): value
                for name, value in model_dump.items()
                if name not in {"point", "stress"}
            },
            "Point": PersistablePoint(X=self.point.x, Z=self.point.z),
            "Stress": self.stress._to_internal_datastructure(),
        }
        return PersistableStatePoint(**data)


class DStabilityStateLinePoint(DStabilityObject):
    """"""

    id: int = -1
    above: DStabilityStress
    below: DStabilityStress
    is_above_and_below_correlated: bool = False
    is_probabilistic: bool = False
    label: str = ""
    x: float

    def _to_internal_datastructure(self) -> PersistableStateLinePoint:
        model_dump = self.model_dump()
        data = {
            **{
                snake_to_camel(name): value
                for name, value in model_dump.items()
                if name not in {"above", "below"}
            },
            "Above": self.above._to_internal_datastructure(),
            "Below": self.below._to_internal_datastructure(),
        }
        return PersistableStateLinePoint(**data)



from pydantic import BaseModel, ConfigDict

from .meta import MetaData

settings = MetaData()


class BaseDataClass(BaseModel):
    """Base class for *all* pydantic classes in GEOLib."""

    model_config = ConfigDict(
        validate_assignment=True,
        arbitrary_types_allowed=True,
        validate_default=True,
        ser_json_inf_nan="constants",
        extra=settings.extra_fields,
    )



import logging
import os
from subprocess import Popen
from typing import List, Optional

import requests
from pydantic import DirectoryPath, HttpUrl
from requests.auth import HTTPBasicAuth

from geolib.errors import CalculationError
from geolib.models import meta
from geolib.models.base_data_class import BaseDataClass
from geolib.models.base_model import BaseModel, output_filename_from_input

logger = logging.getLogger(__name__)
meta = meta.MetaData()


class BaseModelList(BaseDataClass):
    """Hold multiple models that can be executed in parallel.

    Note that all models need to have a unique filename
    otherwise they will overwrite eachother. This also helps with
    identifying them later."""

    models: List[BaseModel]
    errors: List[str] = []

    def execute(
        self,
        calculation_folder: DirectoryPath,
        timeout_in_seconds: int = meta.timeout,
        nprocesses: Optional[int] = os.cpu_count(),
    ) -> "BaseModelList":
        """Execute all models in this class in parallel.

        We split the list to separate folders and call a batch processes on each folder.
        Note that the order of models will change.
        """

        # manual check as remote execution could result in zero models
        if len(self.models) == 0:
            raise ValueError("Can't execute with zero models.")

        lead_model = self.models[0]
        processes = []
        output_models = []
        errors = []

        # Divide the models over n processes and make sure to copy them to prevent aliasing
        split_models = [self.models[i::nprocesses] for i in range(nprocesses)]
        for i, models in enumerate(split_models):
            if len(models) == 0:
                continue
            unique_folder = calculation_folder / str(i)
            unique_folder.mkdir(parents=True, exist_ok=True)

            for model in models:
                fn = unique_folder / model.filename.name
                model.serialize(fn.resolve())

            executable = meta.console_folder / lead_model.default_console_path
            if not executable.exists():
                logger.error(
                    f"Please make sure the `geolib.env` file points to the console folder. GEOLib now can't find it at `{executable}`"
                )
                raise CalculationError(
                    -1, f"Console executable not found at {executable}."
                )

            process = Popen(
                [str(executable)] + lead_model.console_flags + [str(i)],
                cwd=str(calculation_folder.resolve()),
            )
            processes.append(process)

        # Wait for all processes to be done
        for process in processes:
            logger.debug(f"Executed with {process.args}")
            process.wait(timeout=timeout_in_seconds)

        # Iterate over the models
        for models in split_models:
            for model in models:
                model = model.copy(deep=True)  # prevent aliasing
                output_filename = output_filename_from_input(model)
                if output_filename.exists():
                    try:
                        model.parse(output_filename)
                        output_models.append(model)

                    except ValidationError:
                        logger.warning(
                            f"Ouput file generated but parsing of {output_filename.name} failed."
                        )
                        error = model.get_error_context()
                        errors.append(error)
                else:
                    logger.warning(
                        f"Model @ {output_filename.name} failed. Please check the .err file and batchlog.txt in its folder."
                    )
                    error = model.get_error_context()
                    errors.append(error)

        return self.__class__(models=output_models, errors=errors)

    def execute_remote(self, endpoint: HttpUrl) -> "BaseModelList":
        """Execute all models in this class in parallel on a remote endpoint.

        Note that the order of models will change.
        """
        lead_model = self.models[0]

        response = requests.post(
            requests.compat.urljoin(
                endpoint, f"calculate/{lead_model.__class__.__name__.lower()}s"
            ),
            data="[" + ",".join((model.json() for model in self.models)) + "]",
            auth=HTTPBasicAuth(meta.gl_username, meta.gl_password),
        )
        if response.status_code == 200:
            models = response.json()["models"]
            errors = response.json()["errors"]
            stripped_models = []
            for model in models:
                # remove possibly invalid external metadata
                model.get("meta", {}).pop("console_folder", None)
                stripped_models.append(lead_model.__class__(**model))
            return self.__class__(models=stripped_models, errors=errors)
        else:
            raise CalculationError(response.status_code, response.text)



import abc

from geolib.models.base_data_class import BaseDataClass

from .validators import BaseValidator


class BaseModelStructure(BaseDataClass, abc.ABC):
    @property
    def is_valid(self) -> bool:
        """Validates the current model structure."""
        return self.validator().is_valid

    def validator(self) -> BaseValidator:
        """Set the Validator class."""
        return BaseValidator(self)



import logging
from typing import Set

from geolib.models.validators import BaseValidator

logger = logging.getLogger(__name__)


class DStabilityValidator(BaseValidator):
    """Validator for DStability structure.

    Has access to self.ds from parent class.
    Will run all is_valid_ functions to check for validity."""

    def is_valid_stages(self) -> bool:
        """Number of stages should be the same:"""

        stage_count = 0
        for scenario in self.ds.scenarios:
            for _ in scenario.Stages:
                stage_count += 1

        lengths_set = set()
        valid = True
        model_dump = self.ds.model_dump()
        for key, value in (
            (k, v)
            for k, v in model_dump.items()
            if "result" not in k
            and "scenarios" not in k
            and "calculationsettings" not in k
        ):  # Results not required for stage validity.
            if isinstance(value, list):
                lengths_set.add(len(value))
                if len(value) > stage_count:
                    logger.error(
                        f"{self.is_valid_stages.__doc__} {key} has different number of stages: {len(value)}."
                    )
                    valid = False
                    break

        return valid

    def is_valid_layer_loads(self) -> bool:
        """Each layer load must have a consolidation degree for each soil layer"""
        for scenario_index, _ in enumerate(self.ds.scenarios):
            for stage_index, _ in enumerate(self.ds.scenarios[scenario_index].Stages):
                soil_layer_ids: Set[str] = {
                    layer.LayerId
                    for layer in self.ds._get_soil_layers(
                        scenario_index, stage_index
                    ).SoilLayers
                }

                if len(soil_layer_ids) == 0:
                    return True

                layer_load_layer_ids: Set[str] = set()
                for layer_load in self.ds._get_loads(
                    scenario_index, stage_index
                ).LayerLoads:
                    layer_load_layer_ids.add(layer_load.LayerId)

                    consolidation_layer_id_references: Set[str] = set()
                    for consolidation in layer_load.Consolidations:
                        consolidation_layer_id_references.add(consolidation.LayerId)
                        if consolidation.LayerId is None:
                            return False
                    if soil_layer_ids - consolidation_layer_id_references != set(
                        [layer_load.LayerId]
                    ):  # All other soillayer ids are included except the soillayers own id.
                        return False
                if layer_load_layer_ids != soil_layer_ids:
                    return False
        return True



import abc
import re
from enum import Enum
from pathlib import Path
from typing import BinaryIO, Dict, List, Optional, Set, Type, Union

from pydantic import DirectoryPath, FilePath

from geolib.geometry import Point
from geolib.models import BaseDataClass, BaseModel
from geolib.soils import Soil

from ...utils import camel_to_snake, snake_to_camel
from .dgeoflow_parserprovider import DGeoFlowParserProvider
from .internal import (
    CalculationType,
    CalculationTypeEnum,
    DGeoFlowResult,
    DGeoFlowStructure,
    GroundwaterFlowResult,
    InternalPipeTrajectory,
    PersistableSoil,
    PersistableSoilLayer,
    PipeLengthResult,
    Scenario,
    SoilCollection,
    SoilLayerCollection,
)
from .serializer import DGeoFlowInputSerializer, DGeoFlowInputZipSerializer


class DGeoFlowObject(BaseModel, metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def _to_DGeoFlow_sub_structure(self):
        raise NotImplementedError


class DGeoFlowModel(BaseModel):
    """D-Geoflow is software for soft soil piping calculations.

    This model can read, modify and create
    .flox files
    """

    current_scenario: int = -1
    current_scenario_index: int = 0
    current_stage_index: int = 0
    current_calculation_index: int = 0
    datastructure: DGeoFlowStructure = DGeoFlowStructure()
    current_id: int = -1

    def __init__(self, *args, **data) -> None:
        super().__init__(*args, **data)
        self.current_id = self.datastructure.get_unique_id()

    @property
    def parser_provider_type(self) -> Type[DGeoFlowParserProvider]:
        return DGeoFlowParserProvider

    @property
    def default_console_path(self) -> Path:
        return Path("DGeoFlowConsole/D-GeoFlow Console.exe")

    @property
    def custom_console_path(self) -> Path:
        return self.get_meta_property("dgeoflow_console_path")

    @property
    def console_flags_post(self) -> List[str]:
        return [
            str(self.current_scenario_index + 1),
            str(self.current_calculation_index + 1),
        ]

    @property
    def soils(self) -> SoilCollection:
        """Enables easy access to the soil in the internal dict-like datastructure. Also enables edit/delete for individual soils."""
        return self.datastructure.soils

    def get_layer(self, scenario_id: int, layer_id: int) -> PersistableSoilLayer:
        """Enables easy access to the soil in the internal dict-like datastructure. Also enables edit/delete for individual soils."""
        soillayers = self.datastructure.soillayers[scenario_id]
        for layer in soillayers.SoilLayers:
            if int(layer.LayerId) == layer_id:
                return layer
        raise ValueError(f"No soil layer found with layer id {layer_id}")

    def _get_next_id(self) -> int:
        self.current_id += 1
        return self.current_id

    def parse(self, *args, **kwargs):
        super().parse(*args, **kwargs)
        self.current_id = self.datastructure.get_unique_id()

    @property
    def output(self) -> DGeoFlowResult:
        return self.get_result(
            self.current_scenario_index, self.current_calculation_index
        )

    def get_result(self, scenario_index: int, calculation_index: int) -> Dict:
        """
        Returns the results of a scenario. Calculation results are based on analysis type and calculation type.

        Args:
            scenario_index (int): Index of a scenario.
            calculation_index (int): Index of a calculation.

        Returns:
            dict: Dictionary containing the analysis results of the scenario.

        Raises:
            ValueError: No results or calculationsettings available
        """
        return self._get_result_substructure(scenario_index, calculation_index)

    def _get_result_substructure(
        self, scenario_index: int, calculation_index: int
    ) -> DGeoFlowResult:
        if self.datastructure.has_result(scenario_index):
            calculation = self.datastructure.scenarios[scenario_index].Calculations[
                calculation_index
            ]
            result_id = calculation.ResultsId
            calculation_type = calculation.CalculationType

            results = self.datastructure.get_result_substructure(calculation_type)

            for result in results:
                if result.Id == result_id:
                    return result

        raise ValueError(f"No result found for result id {scenario_index}")

    def serialize(self, location: Union[FilePath, DirectoryPath, BinaryIO]):
        """Support serializing to directory while developing for debugging purposes."""
        if isinstance(location, Path) and location.is_dir():
            serializer = DGeoFlowInputSerializer(ds=self.datastructure)
        else:
            serializer = DGeoFlowInputZipSerializer(ds=self.datastructure)
        serializer.write(location)

        if isinstance(location, Path):
            self.filename = location

    def add_soil(self, soil: Soil) -> int:
        """
        Add a new soil to the model. The code must be unique, the id will be generated

        Args:
            soil (Soil): a new soil

        Returns:
            int: id of the added soil
        """
        if self.soils.has_soilcode(soil.code):
            raise ValueError(f"The soil with code {soil.code} is already defined.")

        soil.id = self._get_next_id()
        persistant_soil = self.soils.add_soil(soil)
        return persistant_soil.Id

    def edit_soil(self, code: str, **kwargs: dict) -> PersistableSoil:
        """
        Edit an existing soil with parameter names based on the soil class members

        Args:
            code (str): the code of the soil
            kwargs (dict): the parameters and new values for example 'cohesion=2.0, friction_angle=25.0'

        Returns:
            PersistableSoil: the edited soil
        """
        return self.soils.edit_soil(code, **kwargs)

    def edit_soil_by_name(self, name: str, **kwargs: dict) -> PersistableSoil:
        """
        Edit an existing soil with parameter names based on the soil class members

        Args:
            name (str): the name of the soil
            kwargs (dict): the parameters and new values for example 'cohesion=2.0, friction_angle=25.0'

        Returns:
            PersistableSoil: the edited soil
        """
        return self.soils.edit_soil_by_name(name, **kwargs)

    @property
    def points(self):
        """Enables easy access to the points in the internal dict-like datastructure. Also enables edit/delete for individual points."""

    def add_layer(
        self,
        points: List[Point],
        soil_code: str,
        label: str = "",
        notes: str = "",
        scenario_id: int = None,
    ) -> int:
        """
        Add a soil layer to the model

        Args:
            points (List[Point]): list of Point classes, in clockwise order (non closed simple polygon)
            soil_code (str): code of the soil for this layer
            label (str): label defaults to empty string
            notes (str): notes defaults to empty string
            scenario_id (int): scenario to add to, defaults to 0

        Returns:
            int: id of the added layer
        """
        scenario_id = scenario_id if scenario_id is not None else self.current_scenario

        if not self.datastructure.has_scenario(scenario_id):
            raise IndexError(f"scenario {scenario_id} is not available")
        geometry = self.datastructure.geometries[scenario_id]

        soillayerscollection = self.datastructure.soillayers[scenario_id]

        # do we have this soil code?
        if not self.soils.has_soilcode(soil_code):
            raise ValueError(
                f"The soil with code {soil_code} is not defined in the soil collection."
            )

        # add the layer to the geometry
        # the checks on the validity of the points are done in the PersistableLayer class
        persistable_layer = geometry.add_layer(
            id=str(self._get_next_id()), label=label, points=points, notes=notes
        )

        # add the connection between the layer and the soil to soillayers
        soil = self.soils.get_soil(soil_code)
        soillayerscollection.add_soillayer(layer_id=persistable_layer.Id, soil_id=soil.id)
        return int(persistable_layer.Id)

    def add_meshproperties(
        self,
        element_size: float = 1.0,
        label: str = "",
        scenario_id: int = None,
        layer_id: int = None,
    ) -> int:
        """
        Add a mesh properties to the model

        Args:
            element_size: size of the mesh elements for the discretization, defaults to 1.0
            scenario_id (int): scenario to add to, defaults to 0
            layer_id (int): layer to add to

        Returns:
            int: id of the added meshproperties collection
        """
        scenario_id = scenario_id if scenario_id is not None else self.current_scenario
        meshpropertiescollection = self.datastructure.mesh_properties[scenario_id]

        persistable_layer = self.get_layer(scenario_id, layer_id)
        meshpropertiescollection.add_meshproperty(
            layer_id=persistable_layer.LayerId, element_size=element_size, label=label
        )

        return int(meshpropertiescollection.Id)

    def add_boundary_condition(
        self,
        points: List[Point],
        head_level: float,
        label: str = "",
        notes: str = "",
        scenario_id: int = None,
    ) -> int:
        """
        Add boundary conditions to the model

        Args:
            points (List[Point]): list of Point classes, in clockwise order (non closed simple polygon)
            head_level (float): level of the hydraulic head for the boundary condition
            label (str): label defaults to empty string
            notes (str): notes defaults to empty string
            scenario_id (int): scenario to add to, defaults to 0

        Returns:
            int: id of the boundary conditions collection
        """
        scenario_id = scenario_id if scenario_id is not None else self.current_scenario
        boundary_conditions = self.datastructure.boundary_conditions[scenario_id]

        boundary_condition_id = self._get_next_id()
        boundary_conditions.add_boundary_condition(
            boundary_condition_id, label, notes, points, head_level
        )

        return boundary_condition_id

    @property
    def scenarios(self) -> List[Scenario]:
        return self.datastructure.scenarios

    def add_scenario(
        self, label: str = "Scenario", notes: str = "", set_current: bool = True
    ) -> int:
        """Add a new scenario to the model.

        Args:
            label: Label for the scenario.
            notes: Notes for the scenario.
            set_current: Whether to make the new scenario the current scenario.

        Returns:
            the id of the new stage
        """
        new_id = self._get_next_id()
        new_scenario_id, new_unique_id = self.datastructure.add_default_scenario(
            label, notes, new_id
        )

        if set_current:
            self.current_scenario = new_scenario_id
            self.current_stage_index = 0
            self.current_calculation_index = 0

        self.current_id = new_unique_id
        return new_scenario_id

    def add_stage(
        self,
        scenario_index: Optional[int] = None,
        label: str = "Stage",
        notes: str = "",
        set_current=True,
    ) -> int:
        """Add a new stage to the model at the given scenario index.

        Args:
            scenario_index: The scenario index to add the stage to, defaults to the current scenario.
            label: Label for the stage.
            notes: Notes for the stage.
            set_current: Whether to make the new stage the current stage.

        Returns:
            the id of the new stage
        """
        scenario_index = (
            scenario_index if scenario_index is not None else self.current_scenario
        )

        new_id = self._get_next_id()
        new_stage_index = self.datastructure.add_default_stage(
            scenario_index, label, notes, new_id
        )

        if set_current:
            self.current_stage_index = new_stage_index
        return new_stage_index

    def add_calculation(
        self,
        scenario_index: Optional[int] = None,
        label: str = "Calculation",
        notes: str = "",
        set_current: bool = True,
    ) -> int:
        """Add a new calculation to the model.

        Args:
            scenario_index: The scenario index to add the calculation to, defaults to the current scenario.
            label: Label for the calculation.
            notes: Notes for the calculation.
            set_current: Whether to make the new calculation the current calculation.

        Returns:
            the id of the new stage
        """
        scenario_index = (
            scenario_index if scenario_index is not None else self.current_scenario
        )

        new_calculation_index = self.datastructure.add_default_calculation(
            scenario_index, label, notes
        )

        if set_current:
            self.current_calculation_index = new_calculation_index
        return new_calculation_index

    def set_calculation_type(
        self,
        scenario_index: int = 0,
        calculation_index: int = 0,
        calculation_type: CalculationTypeEnum = CalculationTypeEnum.GROUNDWATER_FLOW,
    ) -> None:
        """
        Sets the calculation type of a calculation.

        Args:
            scenario_index (int): The scenario index to add to, defaults to 0.
            calculation_index (int): The calculation index to add to, defaults to 0.
            calculation_type (CalculationTypeEnum): The calculation type, defaults to GROUNDWATER_FLOW.
        """

        self.datastructure.scenarios[scenario_index].Calculations[
            calculation_index
        ].CalculationType = calculation_type

    def set_pipe_trajectory(
        self,
        scenario_index: int = 0,
        calculation_index: int = 0,
        pipe_trajectory: InternalPipeTrajectory = None,
    ) -> None:
        """
        Sets the pipe trajectory for a calculation.

        Args:
            scenario_index (int): The scenario index to add to, defaults to 0.
            calculation_index (int): The calculation index to add to, defaults to 0.
            pipe_trajectory (PipeTrajectory): The pipe trajectory.
        """

        self.datastructure.scenarios[scenario_index].Calculations[
            calculation_index
        ].PipeTrajectory = pipe_trajectory

    def set_critical_head_boundary_condition(
        self,
        scenario_index: int = 0,
        calculation_index: int = 0,
        boundary_condition_id: int = None,
    ) -> None:
        """
        Sets the critical head boundary condition for a calculation.

        Args:
            scenario_index (int): The scenario index to add to, defaults to 0.
            calculation_index (int): The calculation index to add to, defaults to 0.
            boundary_condition_id (int): The id of the critical head boundary condition.
        """

        self.datastructure.scenarios[scenario_index].Calculations[
            calculation_index
        ].CriticalHeadId = str(boundary_condition_id)

    def set_critical_head_search_parameters(
        self,
        scenario_index: int = 0,
        calculation_index: int = 0,
        minimum_head_level: float = 0,
        maximum_head_level: float = 1,
        step_size: float = 0.1,
    ) -> None:
        """
        Sets the critical head search parameters condition for a calculation.

        Args:
            scenario_index (int): The scenario index to add to, defaults to 0.
            calculation_index (int): The calculation index to add to, defaults to 0.
            minimum_head_level (float): The minimum head level to search at, defaults to 0.
            maximum_head_level (float): The maximum head level to search at, defaults to 1.
            step_size (float): The step size to search with, defaults to 0.1.
        """

        calculation = self.datastructure.scenarios[scenario_index].Calculations[
            calculation_index
        ]

        calculation.CriticalHeadSearchSpace.MinimumHeadLevel = minimum_head_level
        calculation.CriticalHeadSearchSpace.MaximumHeadLevel = maximum_head_level
        calculation.CriticalHeadSearchSpace.StepSize = step_size



# -*- coding: utf-8 -*-

"""
This module contains the primary objects that power GEOLib.
"""
import abc
import logging
from abc import abstractmethod
from pathlib import Path
from subprocess import run
from typing import List, Optional, Type, Union

import requests
from pydantic import DirectoryPath, FilePath, HttpUrl, SerializeAsAny, ValidationError
from requests.auth import HTTPBasicAuth

from geolib.errors import CalculationError
from geolib.models import BaseDataClass
from geolib.models.base_model_structure import BaseModelStructure
from geolib.models.meta import MetaData
from geolib.models.parsers import BaseParserProvider

logger = logging.getLogger(__name__)
meta = MetaData()


class BaseModel(BaseDataClass, abc.ABC):
    filename: Optional[Path] = None
    datastructure: Optional[SerializeAsAny[BaseModelStructure]] = None
    """
    This is the base class for all models in GEOLib.
    
    Note that `datastructure` is a `SerializeAsAny` type, which means that
    the inheriting class is serialized according to its own definition (duck-typing).
    This is needed since Pydantic v2 as the default behavior has changed:
    https://docs.pydantic.dev/latest/concepts/serialization/#subclass-instances-for-fields-of-basemodel-dataclasses-typeddict
    """

    def execute(self, timeout_in_seconds: int = meta.timeout) -> "BaseModel":
        """Execute a Model and wait for `timeout` seconds.

        The model is modified in place if the calculation and parsing
        is successful.
        """
        if self.filename is None:
            raise ValueError("Set filename or serialize first!")
        if not self.filename.exists():
            logger.warning("Serializing before executing.")
            self.serialize(self.filename)

        if self.custom_console_path is not None:
            executable = self.custom_console_path
        else:
            executable = meta.console_folder / self.default_console_path

        if not executable.exists():
            logger.error(
                f"Please make sure the `geolib.env` file points to the console folder. GEOLib now can't find it at `{executable}`"
            )
            raise CalculationError(-1, f"Console executable not found at {executable}.")

        process = run(
            [str(executable)]
            + self.console_flags
            + [str(self.filename.resolve())]
            + self.console_flags_post,
            timeout=timeout_in_seconds,
            cwd=str(self.filename.resolve().parent),
        )
        logger.debug(f"Executed with {process.args}")

        # Successful run
        output_filename = output_filename_from_input(self)
        logger.info(
            f"Checking for {output_filename}, while process exited with {process.returncode}"
        )
        if output_filename.exists():
            try:
                self.parse(output_filename)
                return self  # TODO Figure out whether we should instantiate a new model (parse is a classmethod)
            except ValidationError:
                logger.warning(
                    f"Output file generated but parsing of {output_filename} failed."
                )
                error = self.get_error_context()
                raise CalculationError(process.returncode, error)

        # Unsuccessful run
        else:
            error = self.get_error_context()
            raise CalculationError(
                process.returncode, error + " Path: " + str(output_filename.absolute())
            )

    def execute_remote(self, endpoint: HttpUrl) -> "BaseModel":
        """Execute a Model on a remote endpoint.

        A new model instance is returned.
        """
        response = requests.post(
            requests.compat.urljoin(
                endpoint, f"calculate/{self.__class__.__name__.lower()}"
            ),
            data=self.json(),
            auth=HTTPBasicAuth(meta.gl_username, meta.gl_password),
        )
        if response.status_code == 200:
            data = response.json()
            # remove possibly invalid external metadata
            data.get("meta", {}).pop("console_folder", None)
            return self.__class__(**data)
        else:
            raise CalculationError(response.status_code, response.text)

    def get_error_context(self) -> str:
        err_fn = output_filename_from_input(self, extension=".err")
        batch_fn = self.filename.parent / "Batchlog.txt"
        error = f"{self.filename.name}\n"
        if err_fn.exists():
            with open(err_fn) as f:
                error += f"### {err_fn} ###\n"
                error += f.read()
        elif batch_fn.exists():
            with open(batch_fn) as f:
                error += f"### {batch_fn} ###\n"
                error += f.read()
        else:
            error = f"Couldn't determine source of error for {self.filename.name}."
        return error

    @abstractmethod
    def serialize(
        self, filename: Union[FilePath, DirectoryPath, None]
    ) -> Union[FilePath, DirectoryPath, None]:
        """Serialize model to input file."""

    @property
    def default_console_path(self) -> Path:
        raise NotImplementedError("Implement in concrete classes.")

    @property
    def custom_console_path(self) -> Optional[Path]:
        return None

    @property
    def console_flags(self) -> List[str]:
        return []

    @property
    def console_flags_post(self) -> List[str]:
        return []

    @property
    @abstractmethod
    def parser_provider_type(self) -> Type[BaseParserProvider]:
        """Returns the parser provider type of the current concrete class.

        Raises:
            NotImplementedError: If not implemented in the concrete class.

        Returns:
            Type[BaseParserProvider] -- Concrete parser provider.
        """
        raise NotImplementedError("Implement in concrete classes.")

    def parse(self, filename: FilePath) -> BaseModelStructure:
        """Parse input or outputfile to Model, depending on extension."""
        self.filename = filename
        # self.datastructure = None
        self.datastructure = self.parser_provider_type().parse(filename)
        return self.datastructure

    @property
    def is_valid(self) -> bool:
        """Checks validity and integrity of structure."""
        if self.datastructure is not None:
            return self.datastructure.is_valid
        else:
            logger.warning("No datastructured parsed yet!")
            return False

    @property
    def input(self):
        """Access internal dict-like datastructure of the input."""
        return self.datastructure

    @property
    def output(self):
        """Access internal dict-like datastructure of the output.

        Requires a successful execute.
        """
        return self.datastructure.results

    def get_meta_property(self, key: str) -> Optional[str]:
        """Get a metadata property from the input file."""
        if hasattr(meta, key):
            return meta.__getattribute__(key)
        else:
            return None

    def set_meta_property(self, key: str, value: str) -> None:
        """Set a metadata property from the input file."""
        if hasattr(meta, key):
            meta.__setattr__(key, value)
        else:
            raise ValueError(f"Metadata property {key} does not exist.")


def output_filename_from_input(model: BaseModel, extension: str = None) -> Path:
    if not extension:
        extension = model.parser_provider_type().output_parsers[-1].suffix_list[0]
    return model.filename.with_suffix(extension)
